# RCAIDE/Library/Methods/Powertrain/Propulsors/Turbojet/append_turbojet_conditions.py
# 
# Created:  Jun 2024, M. Clarke  
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  

import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_propulsor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turbojet_conditions(propulsor, segment, energy_conditions, noise_conditions):
    """
    Initializes turbojet operating conditions for a mission segment.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Turbojet
        Turbojet propulsor component with the following attributes:
            - tag : str
                Identifier for the turbojet
            - items : dict
                Dictionary of subcomponents
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where turbojet conditions will be stored
    noise_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Noise conditions container where turbojet noise conditions will be stored
    
    Returns
    -------
    None
    
    Notes
    -----
    This function initializes the necessary data structures for storing turbojet
    operating conditions during a mission segment. It creates zero-filled arrays
    for various performance parameters and recursively calls the append_operating_conditions
    method for each subcomponent of the turbojet.
    
    The function initializes the following parameters: throttle, commanded_thrust_vector_angle, 
    thrust, power, moment, fuel_flow_rate, inputs and outputs containers
    
    It also creates a core_nozzle container in the noise conditions.
    
    **Major Assumptions**
        * All arrays are initialized with zeros
        * Each component has an append_operating_conditions method
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet.compute_turbojet_performance
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet.compute_thurst
    """
    # unpack 
    ones_row          = segment.state.ones_row 
    
    # add propulsor conditions                  
    energy_conditions.propulsors[propulsor.tag]                               = Conditions()  
    energy_conditions.propulsors[propulsor.tag].throttle                      = 0. * ones_row(1)     
    energy_conditions.propulsors[propulsor.tag].commanded_thrust_vector_angle = 0. * ones_row(1)    
    energy_conditions.propulsors[propulsor.tag].thrust                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].power                         = 0. * ones_row(1) 
    energy_conditions.propulsors[propulsor.tag].moment                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].fuel_flow_rate                = 0. * ones_row(1)
    energy_conditions.propulsors[propulsor.tag].inputs                        = Conditions()
    energy_conditions.propulsors[propulsor.tag].outputs                       = Conditions() 
    noise_conditions.propulsors[propulsor.tag]                                = Conditions()  
    noise_conditions.propulsors[propulsor.tag].core_nozzle                    = Conditions()
    
 
    for tag, item in  propulsor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,energy_conditions) 
            for sub_tag, sub_item in  item.items(): 
                if issubclass(type(sub_item), RCAIDE.Library.Components.Component): 
                    sub_item.append_operating_conditions(segment,energy_conditions)     
    return 


# RCAIDE/Library/Missions/Segments/Ground/Takeoff.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# unpack unknowns
# ---------------------------------------------------------------------------------------------------------------------- 
def initialize_conditions(segment):
    """
    Initializes conditions for aircraft takeoff ground roll

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude : float
                Ground altitude [m]
            - velocity_start : float
                Initial ground speed [m/s]
            - velocity_end : float
                Rotation speed [m/s]
            - ground_incline : float
                Runway incline angle [rad]
            - friction_coefficient : float
                Ground friction coefficient [-]
            - state:
                numerics:
                    dimensionless:
                        control_points : array
                            Discretization points [-]
                conditions : Data
                    State conditions container
                unknowns:
                    ground_velocity : array
                        Ground velocity profile [m/s]
                initials : Data, optional
                    Initial conditions from previous segment
            - analyses:
                weights:
                    vehicle:
                        mass_properties:
                            takeoff : float
                                Aircraft takeoff mass [kg]

    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.ground.incline [rad]
            - conditions.ground.friction_coefficient [-]
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.position_vector [m]
            - conditions.weights.total_mass [kg]
    
    Notes
    -----
    This function sets up the initial conditions for a ground takeoff segment with
    acceleration from initial speed to rotation speed. The segment handles ground
    effects, friction, and incline during the takeoff roll.

    **Calculation Process**
        1. Check initial conditions
        2. Initialize velocity profile:
            v = v0 + (vf - v0)*t where:
                - v0 is initial speed
                - vf is rotation speed
                - t is normalized time/distance
        3. Set ground conditions:
            - Friction coefficient
            - Ground incline
        4. Set position and altitude
        5. Set aircraft mass

    **Major Assumptions**
        * Constant friction coefficient
        * Constant ground incline
        * Linear velocity increase
        * Quasi-steady acceleration
        * Small minimum velocity for numerical stability

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """  

    # use the common initialization # unpack inputs
    alt      = segment.altitude 
    v0       = segment.velocity_start
    vf       = segment.velocity_end 
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]   

    if v0  is None: 
        v0 = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # avoid having zero velocity since aero and propulsion models need non-zero Reynolds number
    if v0 == 0.0: v0 = 0.01
    if vf == 0.0: vf = 0.01
    
    # intial and final speed cannot be the same
    if v0 == vf:
        vf = vf + 0.01
        
    # repack
    segment.air_speed_start = v0
    segment.air_speed_end   = vf
    
    initialized_velocity = (vf - v0)*segment.state.numerics.dimensionless.control_points + v0
    
    # Initialize the x velocity unknowns to speed convergence:
    segment.state.unknowns.ground_velocity = initialized_velocity[1:,0]    

    # pack conditions 
    conditions = segment.state.conditions    
    conditions.frames.inertial.velocity_vector[:,0] = initialized_velocity[:,0]
    conditions.ground.incline[:,0]                  = segment.ground_incline
    conditions.ground.friction_coefficient[:,0]     = segment.friction_coefficient   
    conditions.freestream.altitude[:,0]             = alt
    conditions.frames.inertial.position_vector[:,2] = -alt   
    conditions.weights.total_mass[:,0]              = segment.analyses.aerodynamics.vehicle.mass_properties.takeoff
    conditions.frames.inertial.position_vector[:,:] = conditions.frames.inertial.position_vector[0,:][None,:][:,:]


# RCAIDE/Methods/Energy/Propulsors/Turboprop/size_core.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Powertrain.Propulsors.Turboprop          .compute_thrust import compute_thrust 

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  size_core
# ---------------------------------------------------------------------------------------------------------------------- 
def size_core(turboprop, conditions):
    """
    Sizes the core flow for a turboprop engine at the design condition.
    
    Parameters
    ----------
    turboprop : RCAIDE.Library.Components.Propulsors.Turboprop
        Turboprop engine component with the following attributes:
            - tag : str
                Identifier for the turboprop
            - reference_temperature : float
                Reference temperature for mass flow scaling [K]
            - reference_pressure : float
                Reference pressure for mass flow scaling [Pa]
            - design_thrust : float
                Design thrust at the design point [N]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - energy.propulsors[turboprop.tag] : Data
                Turboprop-specific conditions
                    - total_temperature_reference : numpy.ndarray
                        Reference total temperature [K]
                    - total_pressure_reference : numpy.ndarray
                        Reference total pressure [Pa]
                    - throttle : float
                        Throttle setting [0-1]
                    - thrust_specific_fuel_consumption : numpy.ndarray
                        Thrust specific fuel consumption [kg/(N·s)]
                    - non_dimensional_thrust : numpy.ndarray
                        Non-dimensional thrust
    
    Returns
    -------
    None
        Results are stored in the turboprop object:
          - TSFC : float
              Thrust specific fuel consumption [kg/(N·s)]
          - design_mass_flow_rate : float
              Core mass flow rate at design point [kg/s]
          - compressor_nondimensional_massflow : float
              Non-dimensional mass flow parameter [kg·√K/(s·Pa)]
    
    Notes
    -----
    This function determines the core mass flow rate required to produce the design thrust
    at the specified design conditions. It uses the non-dimensional thrust parameter to
    scale the mass flow appropriately.
    
    **Major Assumptions**
        * Perfect gas behavior
        * Turboprop engine with conventional architecture
        * Design point is at maximum throttle (throttle = 1.0)
    
    **Theory**
    The core mass flow rate is calculated from the design thrust and non-dimensional thrust:
    
    .. math::
        \\dot{m}_{core} = \\frac{F_{design} \\cdot \\text{throttle}}{F_{sp}}
    
    The non-dimensional mass flow parameter is then calculated:
    
    .. math::
        \\dot{m}_{hc} = \\frac{\\dot{m}_{core}}{\\sqrt{\\frac{T_{ref}}{T_{t,ref}}} \cdot \\frac{P_{t,ref}}{P_{ref}}}
    
    References
    ----------
    [1] Mattingly, J.D., "Elements of Gas Turbine Propulsion", 2nd Edition, AIAA Education Series, 2005.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turboprop.compute_thrust
    """
    turboprop_conditions = conditions.energy.propulsors[turboprop.tag]
    Tt_ref               = turboprop_conditions.total_temperature_reference  
    Pt_ref               = turboprop_conditions.total_pressure_reference  
    Tref                 = turboprop.reference_temperature
    Pref                 = turboprop.reference_pressure
    
    # Compute nondimensional thrust
    turboprop_conditions.throttle = 1.0
    compute_thrust(turboprop,conditions) 

    # Store results on turboprop data structure 
    TSFC        = turboprop_conditions.thrust_specific_fuel_consumption
    Fsp         = turboprop_conditions.non_dimensional_thrust  
    mdot_core   = turboprop.design_thrust*turboprop_conditions.throttle/(Fsp) 
    mdhc        = mdot_core/ (np.sqrt(Tref/Tt_ref)*(Pt_ref/Pref))   
    
    turboprop.TSFC                                = TSFC
    turboprop.design_mass_flow_rate               = mdot_core 
    turboprop.compressor_nondimensional_massflow  = mdhc
       
    return    



# RCAIDE/Library/Framework/Analyses/Propulsion/Blade_Element_Momentum_Theory_Helmholtz_Wake/wake_model.py
#
# Created:  Jan 2022, R. Erhard
# Modified:  Jun 2024, M. Clarke  
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Library.Components import Wings 
from RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method.extract_wing_collocation_points import extract_wing_collocation_points
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Performance.Blade_Element_Momentum_Theory_Helmholtz_Wake  import compute_wake_induced_velocity
from RCAIDE.Library.Methods.Aerodynamics.Common.Lift.BET_calculations import compute_airfoil_aerodynamics,compute_inflow_and_tip_loss
# Python imports
import numpy as np 
import scipy as sp

# ----------------------------------------------------------------------------------------------------------------------
# wake model
# ---------------------------------------------------------------------------------------------------------------------- 
def evaluate_wake(rotor,wake_inputs,conditions):
    """
    Evaluates the rotor wake using Helmholtz vortex theory.
    
    Parameters
    ----------
    rotor : RCAIDE.Library.Components.Powertrain.Converters.Rotor
        Rotor component with the following attributes:
            - number_of_blades : int
                Number of blades on the rotor
            - tip_radius : float
                Tip radius of the rotor [m]
            - hub_radius : float
                Hub radius of the rotor [m]
            - sol_tolerance : float
                Solution tolerance for wake convergence
    wake_inputs : Data
        Wake input parameters with:
            - ctrl_pts : int
                Number of control points
            - Nr : int
                Number of radial stations
            - Na : int
                Number of azimuthal stations
            - use_2d_analysis : bool
                Flag for 2D (azimuthal) analysis
            - velocity_total : array_like
                Total velocity magnitude [m/s]
            - velocity_axial : array_like
                Axial velocity component [m/s]
            - velocity_tangential : array_like
                Tangential velocity component [m/s]
            - twist_distribution : array_like
                Blade twist distribution [rad]
            - chord_distribution : array_like
                Blade chord distribution [m]
            - radius_distribution : array_like
                Radial station positions [m]
            - speed_of_sounds : array_like
                Speed of sound [m/s]
            - dynamic_viscosities : array_like
                Dynamic viscosity [kg/(m·s)]
    conditions : Data
        Flight conditions
    
    Returns
    -------
    va : array_like
        Axially-induced velocity from rotor wake [m/s]
    vt : array_like
        Tangentially-induced velocity from rotor wake [m/s]
    
    Notes
    -----
    This function evaluates the rotor wake using Helmholtz vortex theory to calculate
    the induced velocities. It solves for the inflow angle (PSI) that satisfies the
    circulation equation, then computes the axial and tangential induced velocities.
    
    The computation follows these steps:
        1. Initialize the inflow angle (PSI) array
        2. Solve for the inflow angle using a nonlinear equation solver
        3. Calculate the axial and tangential induced velocities from the converged solution
    
    **Major Assumptions**
        * The wake is modeled using Helmholtz vortex theory
        * The solution converges to a steady state
        * The inflow angle (PSI) is the primary variable being solved for
    
    **Theory**
    The wake model is based on Helmholtz vortex theory, which relates the circulation
    around the blade to the induced velocities in the wake. The key equation being solved is:
    
    .. math::
        R = \\Gamma - \\frac{1}{2}W \\cdot c \\cdot C_l = 0
    
    where:
        - Γ is the circulation
        - W is the relative velocity
        - c is the chord
        - Cl is the lift coefficient
    
    The circulation is related to the tangential induced velocity by:
    
    .. math::
        \\Gamma = v_t \\cdot \\frac{4\\pi r}{B} \\cdot F \\cdot \\sqrt{1 + \\left(\\frac{4\\lambda_w R}{\\pi B r}\\right)^2}
    
    where:
        - vt is the tangential induced velocity
        - r is the radial position
        - B is the number of blades
        - F is the tip loss factor
        - λw is the inflow ratio
    
    References
    ----------
    [1] Drela, M. "Qprop Formulation", MIT AeroAstro, June 2006 http://web.mit.edu/drela/Public/web/qprop/qprop_theory.pdf
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Performance.Blade_Element_Momentum_Theory_Helmholtz_Wake.BEMT_Helmholtz_performance
    """
    
    va, vt = wake_convergence(rotor, wake_inputs)
        
    return va, vt

def evaluate_slipstream(rotor,rotor_conditions,geometry,ctrl_pts,wing_instance=None):
    """
    Evaluates the velocities induced by the rotor on a specified wing of the vehicle.
    If no wing instance is specified, uses main wing or last available wing in geometry.
    
    Assumptions:
    None

    Source:
    N/A

    Inputs:
       self         - rotor wake
       rotor        - rotor
       geometry     - vehicle geometry
       
    Outputs:
       wake_V_ind   - induced velocity from rotor wake at (VD.XC, VD.YC, VD.ZC)
    
    Properties Used:
    None
    """
    # Check for wing if wing instance is unspecified
    if wing_instance == None:
        nmw = 0
        # check for main wing
        for i,wing in enumerate(geometry.wings):
            if not isinstance(wing,Wings.Main_Wing): continue
            nmw +=1                
            wing_instance = wing
            wing_instance_idx = i
        if nmw == 1:
            pass
        elif nmw>1:
            print("No wing specified for slipstream analysis. Multiple main wings in vehicle, using the last one.")
        else:
            print("No wing specified for slipstream analysis. No main wing defined, using the last wing in vehicle.")
            wing_instance = wing 
            wing_instance_idx = i
    
    # Isolate the VD components corresponding to this wing instance
    wing_CPs, slipstream_vd_ids = extract_wing_collocation_points(geometry, wing_instance_idx)
    
    # Evaluate rotor slipstream effect on specified wing instance
    rot_V_wake_ind = evaluate_wake_velocities(rotor,rotor_conditions,wing_CPs,ctrl_pts)
    
    # Expand
    wake_V_ind = np.zeros((ctrl_pts,geometry.vortex_distribution.n_cp,3))
    wake_V_ind[:,slipstream_vd_ids,:] = rot_V_wake_ind
    
        
    return wake_V_ind

def evaluate_wake_velocities(rotor,rotor_conditions,evaluation_points,ctrl_pts):
    """
    Links the rotor wake to compute the wake-induced velocities at the specified
    evaluation points.
    
    Assumptions:
    None

    Source:
    N/A

    Inputs:
       self               - rotor wake
       rotor              - rotor
       evaluation_points  - points at which to evaluate the rotor wake-induced velocities 
       
    Outputs:
       prop_V_wake_ind  - induced velocity from rotor wake at (VD.XC, VD.YC, VD.ZC)
    
    Properties Used:
    None
    """  
     
    rot_V_wake_ind = compute_wake_induced_velocity(rotor,rotor_conditions,evaluation_points,ctrl_pts)  
    
    return rot_V_wake_ind


# ---------------------------------------------------------------------------------------------------------------------- 
#  wake_convergence
# ---------------------------------------------------------------------------------------------------------------------- 
def wake_convergence(rotor,wake_inputs):
    """
    Wake evaluation is performed using a simplified vortex wake method for Fidelity Zero, 
    following Helmholtz vortex theory.
    
    Assumptions:
    None

    Source:
    Drela, M. "Qprop Formulation", MIT AeroAstro, June 2006
    http://web.mit.edu/drela/Public/web/qprop/qprop_theory.pdf

    Inputs:
       self         - rotor wake
       rotor        - SUAVE rotor
       wake_inputs.
          Ua        - Axial velocity
          Ut        - Tangential velocity
          r         - radius distribution
       
       
    Outputs:
       va  - axially-induced velocity from rotor wake
       vt  - tangentially-induced velocity from rotor wake
    
    Properties Used:
    None
    
    """        
    
    # Unpack some wake inputs
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na    

    if wake_inputs.use_2d_analysis:
        PSI    = np.ones((ctrl_pts,Nr,Na))
    else:
        PSI    = np.ones((ctrl_pts,Nr))

    PSI_final,infodict,ier,msg = sp.optimize.fsolve(iteration,PSI,args=(wake_inputs,rotor),xtol=rotor.sol_tolerance,full_output = 1,band=(1,0))
    
    # Calculate the velocities given PSI
    va, vt = va_vt(PSI_final, wake_inputs, rotor)

    
    return va, vt

def iteration(PSI, wake_inputs, rotor):
    """
    Computes the BEVW iteration.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       B                          number of rotor blades                          [-]
       beta                       blade twist distribution                        [-]
       r                          radius distribution                             [m]
       R                          tip radius                                      [m]
       Wt                         tangential velocity                             [m/s]
       Wa                         axial velocity                                  [m/s]
       U                          total velocity                                  [m/s]
       Ut                         tangential velocity                             [m/s]
       Ua                         axial velocity                                  [m/s]
       cos_psi                    cosine of the inflow angle PSI                  [-]
       sin_psi                    sine of the inflow angle PSI                    [-]
       piece                      output of a step in tip loss calculation        [-]

    Outputs:
       dR_dpsi                    derivative of residual wrt inflow angle         [-]

    """    
    
    # Unpack inputs to rotor wake fidelity zero
    U               = wake_inputs.velocity_total
    Ua              = wake_inputs.velocity_axial
    Ut              = wake_inputs.velocity_tangential
    use_2d_analysis = wake_inputs.use_2d_analysis        
    beta            = wake_inputs.twist_distribution
    c               = wake_inputs.chord_distribution
    r               = wake_inputs.radius_distribution
    a               = wake_inputs.speed_of_sounds
    nu              = wake_inputs.dynamic_viscosities
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na

    # Unpack rotor data        
    R            = rotor.tip_radius
    B            = rotor.number_of_blades
    tc           = rotor.thickness_to_chord
    airfoils     = rotor.airfoils
    a_loc        = rotor.airfoil_polar_stations
    
    # Reshape PSI because the solver gives it flat
    if wake_inputs.use_2d_analysis:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr,Na))
    else:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr))
    
    # compute velocities
    sin_psi      = np.sin(PSI)
    cos_psi      = np.cos(PSI)
    Wa           = 0.5*Ua + 0.5*U*sin_psi
    Wt           = 0.5*Ut + 0.5*U*cos_psi
    vt           = Ut - Wt

    # compute blade airfoil forces and properties
    Cl, Cdval, alpha, alpha_disc,Ma,W,Re,Re_disc = compute_airfoil_aerodynamics(beta,c,r,R,B,Wa,Wt,a,nu,airfoils,a_loc,ctrl_pts,Nr,Na,tc,use_2d_analysis)

    # compute inflow velocity and tip loss factor
    lamdaw, F, piece = compute_inflow_and_tip_loss(r,R,Wa,Wt,B)

    # compute Newton residual on circulation
    Gamma       = vt*(4.*np.pi*r/B)*F*(1.+(4.*lamdaw*R/(np.pi*B*r))*(4.*lamdaw*R/(np.pi*B*r)))**0.5
    Rsquiggly   = Gamma - 0.5*W*c*Cl
    
    return Rsquiggly.flatten()
 
def va_vt(PSI, wake_inputs, rotor):
    """
    Computes the inflow velocities

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       B                          number of rotor blades                          [-]
       beta                       blade twist distribution                        [-]
       r                          radius distribution                             [m]
       R                          tip radius                                      [m]
       Wt                         tangential velocity                             [m/s]
       Wa                         axial velocity                                  [m/s]
       U                          total velocity                                  [m/s]
       Ut                         tangential velocity                             [m/s]
       Ua                         axial velocity                                  [m/s]
       cos_psi                    cosine of the inflow angle PSI                  [-]
       sin_psi                    sine of the inflow angle PSI                    [-]
       piece                      output of a step in tip loss calculation        [-]

    Outputs:
       dR_dpsi                    derivative of residual wrt inflow angle         [-]

    """    
    
    # Unpack inputs to rotor wake fidelity zero
    U               = wake_inputs.velocity_total
    Ua              = wake_inputs.velocity_axial
    Ut              = wake_inputs.velocity_tangential
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na
    
    # Reshape PSI because the solver gives it flat
    if wake_inputs.use_2d_analysis:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr,Na))
    else:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr))
    
    # compute velocities
    sin_psi      = np.sin(PSI)
    cos_psi      = np.cos(PSI)
    Wa           = 0.5*Ua + 0.5*U*sin_psi
    Wt           = 0.5*Ut + 0.5*U*cos_psi
    va           = Wa - Ua
    vt           = Ut - Wt

    return va, vt
 
def compute_dR_dpsi(PSI,wake_inputs,rotor):
    """
    Computes the analytical derivative for the BEVW iteration.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
       B                          number of rotor blades                          [-]
       beta                       blade twist distribution                        [-]
       r                          radius distribution                             [m]
       R                          tip radius                                      [m]
       Wt                         tangential velocity                             [m/s]
       Wa                         axial velocity                                  [m/s]
       U                          total velocity                                  [m/s]
       Ut                         tangential velocity                             [m/s]
       Ua                         axial velocity                                  [m/s]
       cos_psi                    cosine of the inflow angle PSI                  [-]
       sin_psi                    sine of the inflow angle PSI                    [-]
       piece                      output of a step in tip loss calculation        [-]

    Outputs:
       dR_dpsi                    derivative of residual wrt inflow angle         [-]

    """
    # Unpack inputs to rotor wake fidelity zero
    U               = wake_inputs.velocity_total
    Ua              = wake_inputs.velocity_axial
    Ut              = wake_inputs.velocity_tangential
    beta            = wake_inputs.twist_distribution
    r               = wake_inputs.radius_distribution
    ctrl_pts        = wake_inputs.ctrl_pts
    Nr              = wake_inputs.Nr
    Na              = wake_inputs.Na    

    # Unpack rotor data        
    R        = rotor.tip_radius
    B        = rotor.number_of_blades      
    
    # Reshape PSI because the solver gives it flat
    if wake_inputs.use_2d_analysis:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr,Na))
    else:
        PSI    = np.reshape(PSI,(ctrl_pts,Nr))    
    
    
    # An analytical derivative for dR_dpsi used in the Newton iteration for the BEVW
    # This was solved symbolically in Matlab and exported
    # compute velocities
    sin_psi      = np.sin(PSI)
    cos_psi      = np.cos(PSI)
    Wa           = 0.5*Ua + 0.5*U*sin_psi
    Wt           = 0.5*Ut + 0.5*U*cos_psi
    
    lamdaw, F, piece = compute_inflow_and_tip_loss(r,R,Wa,Wt,B)
    
    pi          = np.pi
    pi2         = np.pi**2
    BB          = B*B
    BBB         = BB*B
    f_wt_2      = 4*Wt*Wt
    f_wa_2      = 4*Wa*Wa
    arccos_piece = np.arccos(piece)
    Ucospsi     = U*cos_psi
    Usinpsi     = U*sin_psi
    Utcospsi    = Ut*cos_psi
    Uasinpsi    = Ua*sin_psi
    UapUsinpsi  = (Ua + Usinpsi)
    utpUcospsi  = (Ut + Ucospsi)
    utpUcospsi2 = utpUcospsi*utpUcospsi
    UapUsinpsi2 = UapUsinpsi*UapUsinpsi
    dR_dpsi     = ((4.*U*r*arccos_piece*sin_psi*((16.*UapUsinpsi2)/(BB*pi2*f_wt_2) + 1.)**(0.5))/B -
                   (pi*U*(Ua*cos_psi - Ut*sin_psi)*(beta - np.arctan((Wa+Wa)/(Wt+Wt))))/(2.*(f_wt_2 + f_wa_2)**(0.5))
                   + (pi*U*(f_wt_2 +f_wa_2)**(0.5)*(U + Utcospsi  +  Uasinpsi))/(2.*(f_wa_2/(f_wt_2) + 1.)*utpUcospsi2)
                   - (4.*U*piece*((16.*UapUsinpsi2)/(BB*pi2*f_wt_2) + 1.)**(0.5)*(R - r)*(Ut/2. -
                    (Ucospsi)/2.)*(U + Utcospsi + Uasinpsi ))/(f_wa_2*(1. - np.exp(-(B*(Wt+Wt)*(R -
                    r))/(r*(Wa+Wa))))**(0.5)) + (128.*U*r*arccos_piece*(Wa+Wa)*(Ut/2. - (Ucospsi)/2.)*(U +
                    Utcospsi  + Uasinpsi ))/(BBB*pi2*utpUcospsi*utpUcospsi2*((16.*f_wa_2)/(BB*pi2*f_wt_2) + 1.)**(0.5)))

    dR_dpsi[np.isnan(dR_dpsi)] = 0.1
    
    # This needs to be made into a jacobian
    dR_dpsi = dR_dpsi.flatten()
    L       = np.size(PSI)
    jac     = np.eye(L)*dR_dpsi
    
    return jac


# RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/evaluate_cantera.py

# Created: June 2024, M. Clarke, M. Guidotti 
# Updated: Mar 2025, M. Guidotti, J. Dost, D. Mehta
# Updates: Apr 2025, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from   RCAIDE.Framework.Core import Data  
import numpy                 as np
import os

try:
    import cantera as ct
    CANTERA_AVAILABLE = True
except ImportError:
    ct = None
    CANTERA_AVAILABLE = False

# ----------------------------------------------------------------------------------------------------------------------
#  evaluate_cantera
# ----------------------------------------------------------------------------------------------------------------------   
def evaluate_cantera(combustor,T,P,mdot_air,FAR): 

    """
    Evaluates emission indices using a Chemical Reactor Network (CRN) built in Cantera.
    
    Parameters
    ----------
    combustor : Data
        Combustor configuration data
        - diameter : float
            Combustor diameter [m]
        - length : float
            Combustor length [m]
        - number_of_combustors : int
            Number of combustors for one engine [-]
        - F_SC : float
            Fuel scale factor [-]
        - N_PZ : int
            Number of PSR in the Primary Zone [-]
        - L_PZ : float
            Primary Zone length [m]
        - S_PZ : float
            Mixing parameter in the Primary Zone [-]
        - design_equivalence_ratio_PZ : float
            Design Equivalence Ratio in Primary Zone at Maximum Throttle [-]
        - N_SZ : int
            Number of discretizations in the Secondary Zone [-]
        - f_SM : float
            Slow mode fraction [-]
        - l_SA_SM : float
            Secondary air length fraction (of L_SZ) in slow mode [-]
        - l_SA_FM : float
            Secondary air length fraction (of L_SZ) in fast mode [-]
        - l_DA_start : float
            Dilution air start length fraction (of L_SZ) [-]
        - l_DA_end : float
            Dilution air end length fraction (of L_SZ) [-]
        - joint_mixing_fraction : float
            Joint mixing fraction [-]
        - design_equivalence_ratio_SZ : float
            Design Equivalence Ratio in Secondary Zone [-]
        - air_mass_flow_rate_take_off : float
            Air mass flow rate at take-off [kg/s]
        - fuel_to_air_ratio_take_off : float
            Fuel to air ratio at take-off [-]
        - air_data : Data
            Air object containing air surrogate species
        - fuel_data : Data
            Fuel object containing fuel properties and kinetics
    
    T : float
        Stagnation Temperature entering combustors [K]
    P : float
        Stagnation Pressure entering combustors [Pa]
    mdot_air : float
        Air mass flow entering the combustor [kg/s]
    FAR : float
        Fuel-to-Air ratio [-]
    
    Returns
    -------
    results : Data
        Container for emission indices
        - EI_CO2 : float
            CO2 emission index [kg_CO2/kg_fuel]
        - EI_CO : float
            CO emission index [kg_CO/kg_fuel]
        - EI_H2O : float
            H2O emission index [kg_H2O/kg_fuel]
        - EI_NOx : float
            NOx emission index [kg_NOx/kg_fuel]
        - final_phi : float
            Final equivalence ratio [-]
        - final_T : float
            Final temperature [K]
        - PZ_phi : list
            Equivalence ratio in the Primary Zone [-]
        - PZ_T : list
            Temperature in the Primary Zone [K]
        - PZ_f_psr : list
            Fraction of mass flow entering each PSR [-]
        - PZ_EI_CO2 : list
            CO2 emission index in the Primary Zone [kg_CO2/kg_fuel]
        - PZ_EI_CO : list
            CO emission index in the Primary Zone [kg_CO/kg_fuel]
        - PZ_EI_H2O : list
            H2O emission index in the Primary Zone [kg_H2O/kg_fuel]
        - PZ_EI_NOx : list
            NOx emission index in the Primary Zone [kg_NOx/kg_fuel]
        - SZ_sm_z : list
            Positions in the Secondary Zone slow mode [-]
        - SZ_sm_phi : list
            Equivalence ratio in the Secondary Zone slow mode [-]
        - SZ_sm_T : list
            Temperature in the Secondary Zone slow mode [K]
        - SZ_sm_EI_CO2 : list
            CO2 emission index in the Secondary Zone slow mode [kg_CO2/kg_fuel]
        - SZ_sm_EI_CO : list
            CO emission index in the Secondary Zone slow mode [kg_CO/kg_fuel]
        - SZ_sm_EI_H2O : list
            H2O emission index in the Secondary Zone slow mode [kg_H2O/kg_fuel]
        - SZ_sm_EI_NOx : list
            NOx emission index in the Secondary Zone slow mode [kg_NOx/kg_fuel]
        - SZ_fm_z : list
            Positions in the Secondary Zone fast mode [-]
        - SZ_fm_phi : list
            Equivalence ratio in the Secondary Zone fast mode [-]
        - SZ_fm_T : list
            Temperature in the Secondary Zone fast mode [K]
        - SZ_fm_EI_CO2 : list
            CO2 emission index in the Secondary Zone fast mode [kg_CO2/kg_fuel]
        - SZ_fm_EI_CO : list
            CO emission index in the Secondary Zone fast mode [kg_CO/kg_fuel]
        - SZ_fm_EI_H2O : list
            H2O emission index in the Secondary Zone fast mode [kg_H2O/kg_fuel]
        - SZ_fm_EI_NOx : list
            NOx emission index in the Secondary Zone fast mode [kg_NOx/kg_fuel]
        - SZ_joint_z : list
            Positions in the Secondary Zone joint mode [-]
        - SZ_joint_phi : list
            Equivalence ratio in the Secondary Zone joint mode [-]
        - SZ_joint_T : list
            Temperature in the Secondary Zone joint mode [K]
        - SZ_joint_EI_CO2 : list
            CO2 emission index in the Secondary Zone joint mode [kg_CO2/kg_fuel]
        - SZ_joint_EI_CO : list
            CO emission index in the Secondary Zone joint mode [kg_CO/kg_fuel]
        - SZ_joint_EI_H2O : list
            H2O emission index in the Secondary Zone joint mode [kg_H2O/kg_fuel]
        - SZ_joint_EI_NOx : list
            NOx emission index in the Secondary Zone joint mode [kg_NOx/kg_fuel]
    
    Notes
    -----
    This function uses Cantera to simulate the chemical kinetics and thermodynamics of the combustor. It requires the Cantera module to be installed.
    
    **Extra modules required**
    
    * Cantera
    
    **Major Assumptions**
        * The combustor operates under steady-state conditions.
        * The fuel and air are perfectly mixed before entering the combustor.
    
    **Theory**
    
    The function evaluates the emission indices by simulating the chemical reactions in the combustor using Cantera. The emissions are calculated based on the chemical kinetics and thermodynamics of the fuel and air mixture.
    
    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics, thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    [2] Brink, L. F. J. (2020). Modeling the impact of fuel composition on aircraft engine NOₓ, CO, and soot emissions. Master's thesis, Massachusetts Institute of Technology.
    [3] Allaire, D. L. (2006). A physics-based emissions model for aircraft gas turbine combustors. Master's thesis, Massachusetts Institute of Technology.
    [4] Lefebvre, A. H., & Ballal, D. R. (2010). Gas turbine combustion: Alternative fuels and emissions (3rd ed.). CRC Press.
    
    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_CRN_emission_indices_no_surrogate
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_CRN_emission_indices_surrogate
    """
    
    # ------------------------------------------------------------------------------              
    # ------------------------------ Combustor Inputs ------------------------------              
    # ------------------------------------------------------------------------------              

    rcaide_root    = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
    mechanism_path = os.path.join(rcaide_root, 'Emissions', 'Chemical_Reactor_Network_Method', 'Data')
    gas            = mechanism_path + '/' + combustor.fuel_data.kinetic_mechanism

    data                  = Data()         
    data.final            = Data()
    data.final.EI         = Data()
    data.final.EI.CO2     = 0 # [kg/kg_fuel]                             
    data.final.EI.CO      = 0 # [kg/kg_fuel]                               
    data.final.EI.H2O     = 0 # [kg/kg_fuel]                               
    data.final.EI.NOx     = 0 # [kg/kg_fuel]  
    data.final.phi        = 0 # [-] 
    data.final.T          = 0 # [K] 
    data.PZ               = Data()
    data.PZ.psr           = Data()
    data.PZ.psr.phi       = [] # [-] 
    data.PZ.psr.T         = [] # [K] 
    data.PZ.psr.f_psr     = [] # [-] 
    data.PZ.psr.EI        = Data()
    data.PZ.psr.EI.CO2    = [] # [kg/kg_fuel] 
    data.PZ.psr.EI.CO     = [] # [kg/kg_fuel] 
    data.PZ.psr.EI.H2O    = [] # [kg/kg_fuel] 
    data.PZ.psr.EI.NOx    = [] # [kg/kg_fuel] 
    data.PZ.phi           = 0 # [-] 
    data.PZ.T             = 0 # [K] 
    data.PZ.f_psr         = 0 # [-] 
    data.PZ.EI            = Data()
    data.PZ.EI.CO2        = 0 # [kg/kg_fuel] 
    data.PZ.EI.CO         = 0 # [kg/kg_fuel] 
    data.PZ.EI.H2O        = 0 # [kg/kg_fuel] 
    data.PZ.EI.NOx        = 0 # [kg/kg_fuel] 
    data.SZ               = Data()
    data.SZ.sm            = Data()
    data.SZ.sm.z          = [] # [-] 
    data.SZ.sm.phi        = [] # [-] 
    data.SZ.sm.T          = [] # [K] 
    data.SZ.sm.EI         = Data()
    data.SZ.sm.EI.CO2     = [] # [kg/kg_fuel] 
    data.SZ.sm.EI.CO      = [] # [kg/kg_fuel] 
    data.SZ.sm.EI.H2O     = [] # [kg/kg_fuel] 
    data.SZ.sm.EI.NOx     = [] # [kg/kg_fuel] 
    data.SZ.fm            = Data()
    data.SZ.fm.z          = [] # [-] 
    data.SZ.fm.phi        = [] # [-] 
    data.SZ.fm.T          = [] # [K] 
    data.SZ.fm.EI         = Data()
    data.SZ.fm.EI.CO2     = [] # [kg/kg_fuel] 
    data.SZ.fm.EI.CO      = [] # [kg/kg_fuel] 
    data.SZ.fm.EI.H2O     = [] # [kg/kg_fuel] 
    data.SZ.fm.EI.NOx     = [] # [kg/kg_fuel] 
    data.SZ.joint         = Data()
    data.SZ.joint.z       = [] # [-] 
    data.SZ.joint.phi     = [] # [-] 
    data.SZ.joint.T       = [] # [K] 
    data.SZ.joint.EI      = Data()
    data.SZ.joint.EI.CO2  = [] # [kg/kg_fuel] 
    data.SZ.joint.EI.CO   = [] # [kg/kg_fuel] 
    data.SZ.joint.EI.H2O  = [] # [kg/kg_fuel] 
    data.SZ.joint.EI.NOx  = [] # [kg/kg_fuel]  

    
    compute_combustor_performance(data, combustor, T, P, mdot_air, FAR, gas) # [-] Run combustor function                                                           
     
    return data

# ----------------------------------------------------------------------
#  RQL Burner Model
# ----------------------------------------------------------------------

def compute_combustor_performance(results, combustor, Temp_air, Pres_air, mdot_air_tot, FAR, gas):
    if CANTERA_AVAILABLE:
        mdot_fuel_TakeOff = combustor.fuel_to_air_ratio_take_off * combustor.air_mass_flow_rate_take_off # [kg/s] Fuel mass flow rate at Take Off
        mdot_fuel_tot     = mdot_air_tot * FAR                             # [kg/s] Fuel mass flow rate 
        mdot_air          = mdot_air_tot / combustor.number_of_combustors  # [kg/s] Air mass flow rate per combustor
        mdot_fuel         = mdot_fuel_tot / combustor.number_of_combustors # [kg/s] Fuel mass flow rate per combustor

        f_air_PZ          = mdot_fuel_TakeOff * combustor.F_SC / (combustor.design_equivalence_ratio_PZ * combustor.air_mass_flow_rate_take_off * combustor.fuel_data.stoichiometric_fuel_air_ratio) # [-] Air mass flow rate fraction in Primary Zone
        phi_sign          = (mdot_fuel * combustor.F_SC) / (mdot_air * f_air_PZ * combustor.fuel_data.stoichiometric_fuel_air_ratio) # [-] Mean equivalence ratio in the Primary Zone
        sigma_phi         = phi_sign * combustor.S_PZ                      # [-] Standard deviation of the Equivalence Ratio in the Primary Zone
        V_PZ              = (combustor.volume/combustor.length) * combustor.L_PZ # [m^3] Volume of the Primary Zone
        V_PZ_PSR          = V_PZ / combustor.N_PZ                          # [m^3] Volume of each PSR
        mdot_air_PZ       = f_air_PZ * mdot_air                            # [kg/s] Air mass flow rate in the Primary Zone
        phi_PSR           = np.linspace(phi_sign - 2 * sigma_phi, phi_sign + 2 * sigma_phi, combustor.N_PZ) # [-] Equivalence ratio in each PSR 
        Delta_phi         = np.abs(phi_PSR[0] - phi_PSR[1])                # [-] Equivalence ratio step in the Primary Zone

        fuel              = ct.Solution(gas)                               # [-] Fuel object
        fuel.TPX          = combustor.fuel_data.temperature, combustor.fuel_data.pressure, combustor.fuel_data.fuel_surrogate_S1                   # [K, Pa, -] Temperauture, Pressure and Mole fraction composition of fuel
        fuel_reservoir    = ct.Reservoir(fuel)                             # [-] Fuel reservoir
        air               = ct.Solution(gas)                               # [-] Air object
        air.TPX           = Temp_air, Pres_air, combustor.air_data.air_surrogate                      # [K, Pa, -] Temperauture, Pressure and Mole fraction composition of air
        air_reservoir     = ct.Reservoir(air)                              # [-] Air reservoir
        fuel_hot          = ct.Solution(gas)                               # [-] Fuel hot state
        fuel_hot.TPX      = Temp_air, Pres_air, combustor.fuel_data.fuel_surrogate_S1                     # [K, Pa, -] Temperauture, Pressure and Mole fraction composition of hot fuel
        delta_h           = np.abs(fuel.h - fuel_hot.h)                    # [J/kg] Fuel specific enthalpy difference

        PZ_Structures     = {"PSRs": {}, "MFC_AirToPSR": {}, "MFC_FuelToPSR": {}, "PSR_Networks": {}, "MFC_PSRToMixer": {}} # [-] Primary Zone structures
        mdot_PZ, f_PSR_data, mass_psr_list = [], [], [] # [-] Arrays to store results
        mixer             = ct.ConstPressureReactor(air)                   # [-] Mixer object

        phi_diff          = phi_PSR - phi_sign                             # [-] Equivalence ratio difference
        f_PSR_data        = (1 / (np.sqrt(2 * np.pi) * sigma_phi)) * np.exp(-(phi_diff ** 2) / (2 * sigma_phi ** 2)) * Delta_phi # [-] Fraction of mass flow entering each reactor
        f_PSR_data       /= np.sum(f_PSR_data)                             # [-] Normalizes mass flow rate fraction in each PSR
        
        # ------------------------------------------------------------------
        #  Primary Zone (PZ)
        # ------------------------------------------------------------------

        for i in range(combustor.N_PZ):

            f_PSR_PZ_i      = f_PSR_data[i]                                # [-] Fuel mass flow rate fraction in the PSR
            mdot_air_PZ_i   = f_PSR_PZ_i * (mdot_air_PZ + mdot_fuel) / (1 + phi_PSR[i] * combustor.fuel_data.stoichiometric_fuel_air_ratio) # [kg/s] Air mass flow rate in the PSR
            mdot_fuel_PZ_i  = mdot_air_PZ_i * phi_PSR[i] * combustor.fuel_data.stoichiometric_fuel_air_ratio # [kg/s] Fuel mass flow rate in the PSR
            mdot_total_PZ_i = mdot_air_PZ_i + mdot_fuel_PZ_i               # [kg/s] Total mass flow rate in the PSR
            mdot_PZ.append(mdot_total_PZ_i)                                # [-] Store total mass flow rate in the PSR

            h_mix_PZ_i      = (1 / (mdot_air_PZ_i + mdot_fuel_PZ_i)) * (mdot_air_PZ_i * air.h + mdot_fuel_PZ_i * fuel_hot.h - mdot_fuel_PZ_i * (combustor.fuel_data.heat_of_vaporization + delta_h)) # [J/kg] Mixture specific enthalpy
            psr_gas_PZ_i    = ct.Solution(gas)                             # [-] PSR gas object
            psr_gas_PZ_i.set_equivalence_ratio(phi_PSR[i], combustor.fuel_data.fuel_surrogate_S1, combustor.air_data.air_surrogate)  # [-] Set equivalence ratio, fuel, and air mole fractions
            psr_gas_PZ_i.HP = h_mix_PZ_i, Pres_air                         # [J/kg, Pa] Set enthalpy and pressure
            psr_gas_PZ_i.equilibrate('HP')                                 # [-] Equilibrate the gas at constant enthalpy and pressure
            
            psr_PZ_i        = ct.ConstPressureReactor(psr_gas_PZ_i, name=f'PSR_{i+1}') # [-] PSR object
            psr_PZ_i.volume = V_PZ_PSR                                     # [m^3] PSR volume
            PZ_Structures["PSRs"][f'PSR_{i+1}'] = psr_PZ_i                 # [-] Store PSR object

            mfc_air_PZ_i       = ct.MassFlowController(air_reservoir, psr_PZ_i, name=f'AirToPSR_{i+1}', mdot=mdot_air_PZ_i) # [-] Air mass flow controller
            PZ_Structures["MFC_AirToPSR"][f'AirToPSR_{i+1}'] = mfc_air_PZ_i # [-] Store air mass flow controller
            mfc_fuel_PZ_i      = ct.MassFlowController(fuel_reservoir, psr_PZ_i, name=f'FuelToPSR_{i+1}', mdot=mdot_fuel_PZ_i) # [-] Fuel mass flow controller
            PZ_Structures["MFC_FuelToPSR"][f'FuelToPSR_{i+1}'] = mfc_fuel_PZ_i # [-] Store fuel mass flow controller

            psr_network_PZ_i   = ct.ReactorNet([psr_PZ_i])                 # [-] PSR network setup
            rho_PZ_i           = psr_gas_PZ_i.density                      # [kg/m^3] Gas density in the PSR
            t_res_PZ_i         = V_PZ_PSR * rho_PZ_i / mdot_total_PZ_i     # [s] Residence time in the PSR
            mass_psr_list.append(t_res_PZ_i * mdot_total_PZ_i)             # [-] Mass of PSR
            psr_network_PZ_i.advance(t_res_PZ_i)                           # [-] Advance the PSR network

            mfc_out_PZ_i       = ct.MassFlowController(psr_PZ_i, mixer, name=f'PSRToMixer_{i+1}', mdot = mdot_total_PZ_i) # [-] PSR to mixer mass flow controller
            PZ_Structures["MFC_PSRToMixer"][f'PSRToMixer_{i+1}'] = mfc_out_PZ_i # [-] Store PSR to mixer mass flow controller
            
            EI = calculate_emission_indices(psr_PZ_i, mdot_total_PZ_i, mdot_fuel_PZ_i) # [-] Emission indices computation

            results.PZ.psr.phi.append(phi_PSR[i])                 # [-] Store Equivalence ratio
            results.PZ.psr.T.append(psr_gas_PZ_i.T)               # [K] Store Temperature
            results.PZ.psr.f_psr.append(f_PSR_PZ_i)               # [-] Store mass flow rate fraction
            results.PZ.psr.EI.NOx.append(EI['NOx'])            # [kg/kg_fuel] Store NOx emission index
            results.PZ.psr.EI.CO2.append(EI['CO2'])            # [kg/kg_fuel] Store CO2 emission index
            results.PZ.psr.EI.CO.append(EI['CO'])              # [kg/kg_fuel] Store CO emission index
            results.PZ.psr.EI.H2O.append(EI['H2O'])            # [kg/kg_fuel] Store H2O emission index\

        # ----------------------------------------------------------------------
        #  Initial Mixing
        # ----------------------------------------------------------------------
        
        mdot_tot_PZ       = sum(mdot_PZ)                                   # [kg/s] Total mass flow rate of the PZ
        mixture_list      = []                                             # [-] Mixture list
        for i in range(combustor.N_PZ):
            psr_output    = PZ_Structures["PSRs"][f'PSR_{i+1}'].thermo     # [-] PSR output
            mixture       = ct.Quantity(psr_output, constant='HP')         # [-] Mixture Quantity setup
            mixture.TPX   = psr_output.T, psr_output.P, psr_output.X       # [K, Pa, -] Temperauture, Pressure and Mole fraction composition of the ixture
            mixture.moles = mass_psr_list[i] / psr_output.mean_molecular_weight # [-] Mixture moles
            mixture_list.append(mixture)                                   # [-] Store mixture moles
        mixture_sum       = mixture_list[0]                                # [-] Define Mixture sum
        
        for mixture in mixture_list[1:]: 
            mixture_sum  += mixture                                        # [-] Add all into a Mixture sum
    
        EI_mixer_initial = calculate_emission_indices(mixture_sum, mdot_tot_PZ, mdot_fuel) # [-] Emission indices computation

        results.PZ.phi = mixture_sum.equivalence_ratio()    # [-] Store Equivalence ratio
        results.PZ.T = mixture_sum.T                        # [K] Store Temperature
        results.PZ.EI.NOx = EI_mixer_initial['NOx']      # [kg/kg_fuel] Store NOx emission index
        results.PZ.EI.CO2 = EI_mixer_initial['CO2']      # [kg/kg_fuel] Store CO2 emission index
        results.PZ.EI.CO = EI_mixer_initial['CO']        # [kg/kg_fuel] Store CO emission index
        results.PZ.EI.H2O = EI_mixer_initial['H2O']      # [kg/kg_fuel] Store H2O emission index

        # ----------------------------------------------------------------------
        #  Secondary Zone
        # ----------------------------------------------------------------------

        combustor.L_SZ                      = combustor.length - combustor.L_PZ # [m] Secondary Zone length
        A_SZ                                = (combustor.volume/combustor.length)  # [m^2] Cross-sectional area of Secondary Zone
        f_air_SA                            = mdot_fuel_TakeOff / (combustor.design_equivalence_ratio_SZ * combustor.fuel_data.stoichiometric_fuel_air_ratio * combustor.air_mass_flow_rate_take_off) # [-] Secondary air mass flow fraction
        f_air_DA                            = 1 - f_air_PZ - f_air_SA      # [-] Dilution air mass flow fraction
        f_FM                                = 1 - combustor.f_SM           # [-] Fast mode fraction
        beta_SA_FM                          = (f_air_SA * f_FM * mdot_air) / (combustor.l_SA_FM * combustor.L_SZ) # [kg/s/m] Secondary air mass flow rate per unit length in fast mode
        beta_SA_SM                          = (f_air_SA * combustor.f_SM * mdot_air) / (combustor.l_SA_SM * combustor.L_SZ) # [kg/s/m] Secondary air mass flow rate per unit length in slow mode
        beta_DA                             = (f_air_DA * mdot_air) / ((combustor.l_DA_end - combustor.l_DA_start) * combustor.L_SZ) # [kg/s/m] Dilution air mass flow rate per unit length
        mdot_total_sm                       = combustor.f_SM * mdot_tot_PZ # [kg/s] Initial total mass flow rate in slow mode
        mdot_total_fm                       = f_FM * mdot_tot_PZ           # [kg/s] Initial total mass flow rate in fast mode
        dz                                  = combustor.L_SZ / combustor.N_SZ # [m] Discretization step size
        z_positions                         = np.linspace(0, combustor.L_SZ, combustor.N_SZ + 1) # [m] Axial position array

        # Slow Mode 
        mixed_gas_sm                        = ct.Solution(gas)             # [-] Slow mode gas object
        mixed_gas_sm.TPX                    = mixture_sum.T, mixture_sum.P, mixture_sum.X # [K, Pa, -] Initial state from PZ mixture
        reactor_sm                          = ct.ConstPressureReactor(mixed_gas_sm) # [-] Slow mode reactor
        sim_sm                              = ct.ReactorNet([reactor_sm])  # [-] Slow mode reactor network

        for z_sm in z_positions[1:int(combustor.joint_mixing_fraction * combustor.N_SZ) + 1]:
            z_frac_sm                       = z_sm / combustor.L_SZ        # [-] Fractional position in SZ
            mdot_air_added_sm               = (beta_SA_SM * dz if z_frac_sm <= combustor.l_SA_SM else
                                            beta_DA * dz if combustor.l_DA_start <= z_frac_sm <= combustor.l_DA_end else 0.0) # [kg/s] Air mass flow rate added
            previous_mdot_total_sm          = mdot_total_sm                # [kg/s] Previous total mass flow rate
            mdot_total_sm                  += mdot_air_added_sm            # [kg/s] Update total mass flow rate
            residence_time                  = dz * A_SZ * mixed_gas_sm.density / mdot_total_sm # [s] Residence time

            air_qty                         = ct.Quantity(air)             # [-] Air quantity object
            air_qty.mass                    = mdot_air_added_sm * residence_time # [kg] Mass of air added over residence time
            mix_qty                         = ct.Quantity(mixed_gas_sm)    # [-] Mixture quantity object
            mix_qty.mass                    = previous_mdot_total_sm * residence_time # [kg] Mass of existing mixture over residence time
            mixture_sm                      = mix_qty + air_qty            # [-] Combined mixture

            mixed_gas_sm.TP                 = mixture_sm.T, mixture_sm.P   # [K, Pa] Update temperature and pressure
            mixed_gas_sm.Y                  = mixture_sm.Y                 # [-] Update composition

            reactor_sm                      = ct.ConstPressureReactor(mixed_gas_sm) # [-] Slow mode reactor
            sim_sm                          = ct.ReactorNet([reactor_sm])  # [-] Reactor setup
            sim_sm.advance(residence_time)                                 # [-] Advance simulation

            EI_sm                           = calculate_emission_indices(mixed_gas_sm, mdot_total_sm, combustor.f_SM * mdot_fuel) # [-] Emission indices    

            results.SZ.sm.phi.append(mixed_gas_sm.equivalence_ratio()) # [-] Store equivalence ratio
            results.SZ.sm.T.append(mixed_gas_sm.T)                # [K] Store temperature 
            results.SZ.sm.z.append(z_frac_sm * 100)               # [%] Store position percentage 
            results.SZ.sm.EI.CO2.append(EI_sm['CO2'])          # [kg/kg_fuel] Store CO2 emission index 
            results.SZ.sm.EI.NOx.append(EI_sm['NOx'])          # [kg/kg_fuel] Store NOx emission index 
            results.SZ.sm.EI.CO.append(EI_sm['CO'])            # [kg/kg_fuel] Store CO emission index 
            results.SZ.sm.EI.H2O.append(EI_sm['H2O'])          # [kg/kg_fuel] Store H2O emission index

        # Fast Mode
        mixed_gas_fm                        = ct.Solution(gas)             # [-] Fast mode gas object
        mixed_gas_fm.TPX                    = mixture_sum.T, mixture_sum.P, mixture_sum.X # [K, Pa, -] Initial state from PZ mixture
        reactor_fm                          = ct.ConstPressureReactor(mixed_gas_fm) # [-] Fast mode reactor
        sim_fm                              = ct.ReactorNet([reactor_fm])  # [-] Fast mode reactor network

        for z_fm in z_positions[1:int(combustor.joint_mixing_fraction * combustor.N_SZ) + 1]:
            z_frac_fm                       = z_fm / combustor.L_SZ        # [-] Fractional position in SZ
            mdot_air_added_fm               = (beta_SA_FM * dz if z_frac_fm <= combustor.l_SA_FM else
                                            beta_DA * dz if combustor.l_DA_start <= z_frac_fm <= combustor.l_DA_end else 0.0) # [kg/s] Air mass flow rate added
            previous_mdot_total_fm          = mdot_total_fm                # [kg/s] Previous total mass flow rate
            mdot_total_fm                  += mdot_air_added_fm            # [kg/s] Update total mass flow rate
            residence_time                  = dz * A_SZ * mixed_gas_fm.density / mdot_total_fm # [s] Residence time

            air_qty                         = ct.Quantity(air)             # [-] Air quantity object
            air_qty.mass                    = mdot_air_added_fm * residence_time # [kg] Mass of air added over residence time
            mix_qty                         = ct.Quantity(mixed_gas_fm)    # [-] Mixture quantity object
            mix_qty.mass                    = previous_mdot_total_fm * residence_time # [kg] Mass of existing mixture over residence time
            mixture_fm                      = mix_qty + air_qty            # [-] Combined mixture

            mixed_gas_fm.TP                 = mixture_fm.T, mixture_fm.P   # [K, Pa] Update temperature and pressure
            mixed_gas_fm.Y                  = mixture_fm.Y                 # [-] Update composition

            reactor_fm                      = ct.ConstPressureReactor(mixed_gas_fm) # [-] Slow mode reactor
            sim_fm                          = ct.ReactorNet([reactor_fm])  # [-] Reactor setup
            sim_fm.advance(residence_time)                                 # [-] Advance simulation

            EI_fm                           = calculate_emission_indices(mixed_gas_fm, mdot_total_fm, f_FM * mdot_fuel) # [-] Emission indices
            
            results.SZ.fm.phi.append(mixed_gas_fm.equivalence_ratio()) # [-] Store equivalence ratio
            results.SZ.fm.T.append(mixed_gas_fm.T)                # [K] Store temperature 
            results.SZ.fm.z.append(z_frac_fm * 100)               # [%] Store position percentage 
            results.SZ.fm.EI.CO2.append(EI_fm['CO2'])          # [kg/kg_fuel] Store CO2 emission index 
            results.SZ.fm.EI.NOx.append(EI_fm['NOx'])          # [kg/kg_fuel] Store NOx emission index 
            results.SZ.fm.EI.CO.append(EI_fm['CO'])            # [kg/kg_fuel] Store CO emission index 
            results.SZ.fm.EI.H2O.append(EI_fm['H2O'])          # [kg/kg_fuel] Store H2O emission index

        # Joint Mixing 
        mixed_gas_joint                     = ct.Solution(gas)             # [-] Joint mode gas object
        total_mass_flow                     = mdot_total_sm + mdot_total_fm # [kg/s] Total mass flow rate after slow and fast modes
        sm_qty                              = ct.Quantity(mixed_gas_sm)    # [-] Slow mode quantity
        fm_qty                              = ct.Quantity(mixed_gas_fm)    # [-] Fast mode quantity
        joint_mixture                       = sm_qty + fm_qty              # [-] Combined mixture
        mixed_gas_joint.TP                  = joint_mixture.T, joint_mixture.P # [K, Pa] Initial temperature and pressure
        mixed_gas_joint.Y                   = joint_mixture.Y              # [-] Initial composition
        mdot_total_joint                    = total_mass_flow              # [kg/s] Initial total mass flow rate
        reactor_joint                       = ct.ConstPressureReactor(mixed_gas_joint) # [-] Joint mode reactor
        sim_joint                           = ct.ReactorNet([reactor_joint]) # [-] Joint mode reactor network

        for z_joint in z_positions[int(combustor.joint_mixing_fraction * combustor.N_SZ) + 1:]:
            z_frac_joint                    = z_joint / combustor.L_SZ     # [-] Fractional position in SZ
            mdot_air_added_joint            = (beta_SA_FM * dz if z_frac_joint <= combustor.l_SA_FM else
                                            beta_DA * dz if combustor.l_DA_start <= z_frac_joint <= combustor.l_DA_end else 0.0) # [kg/s] Air mass flow rate added
            previous_mdot_total_joint       = mdot_total_joint             # [kg/s] Previous total mass flow rate
            mdot_total_joint               += mdot_air_added_joint         # [kg/s] Update total mass flow rate
            residence_time                  = dz * A_SZ * mixed_gas_joint.density / mdot_total_joint # [s] Residence time

            air_qty                         = ct.Quantity(air)             # [-] Air quantity object
            air_qty.mass                    = mdot_air_added_joint * residence_time # [kg] Mass of air added over residence time
            mix_qty                         = ct.Quantity(mixed_gas_joint) # [-] Mixture quantity object
            mix_qty.mass                    = previous_mdot_total_joint * residence_time # [kg] Mass of existing mixture over residence time
            mixture_joint                   = mix_qty + air_qty            # [-] Combined mixture

            mixed_gas_joint.TP              = mixture_joint.T, mixture_joint.P # [K, Pa] Update temperature and pressure
            mixed_gas_joint.Y               = mixture_joint.Y              # [-] Update composition

            reactor_joint                   = ct.ConstPressureReactor(mixed_gas_joint) # [-] Joint mode reactor
            sim_joint                       = ct.ReactorNet([reactor_joint]) # [-] Reactor setup
            sim_joint.advance(residence_time)                              # [-] Advance simulation

            EI_joint                        = calculate_emission_indices(mixed_gas_joint, mdot_total_joint, mdot_fuel) # [-] Emission indices
            
            results.SZ.joint.phi.append(mixed_gas_joint.equivalence_ratio()) # [-] Store equivalence ratio
            results.SZ.joint.T.append(mixed_gas_joint.T)          # [K] Store temperature 
            results.SZ.joint.z.append(z_frac_joint * 100)         # [%] Store position percentage 
            results.SZ.joint.EI.CO2.append(EI_joint['CO2'])    # [kg/kg_fuel] Store CO2 emission index 
            results.SZ.joint.EI.NOx.append(EI_joint['NOx'])    # [kg/kg_fuel] Store NOx emission index 
            results.SZ.joint.EI.CO.append(EI_joint['CO'])      # [kg/kg_fuel] Store CO emission index 
            results.SZ.joint.EI.H2O.append(EI_joint['H2O'])    # [kg/kg_fuel] Store H2O emission index

        results.final.phi = mixed_gas_joint.equivalence_ratio() # [-] Store Equivalence ratio
        results.final.T = mixed_gas_joint.T                    # [K] Store Temperature
        results.final.EI.NOx = EI_joint['NOx']              # [kg/kg_fuel] Store NOx emission index
        results.final.EI.CO2 = EI_joint['CO2']              # [kg/kg_fuel] Store CO2 emission index
        results.final.EI.CO = EI_joint['CO']                # [kg/kg_fuel] Store CO emission index
        results.final.EI.H2O = EI_joint['H2O']              # [kg/kg_fuel] Store H2O emission index

    return results

def calculate_emission_indices(reactor,  mdot_total, mdot_fuel):
    """Calculate emission indices for combustion products"""
    gas                                 = reactor.thermo if hasattr(reactor, 'thermo') else reactor # [-] Extract gas object
    NOx_species                         = ['NO', 'NO2']              # [-] List of NOx species
    EI_NOx                              = 0.0                        # [kg/kg_fuel] Initialize NOx emission index
    for species in NOx_species:
        try:
            idx                         = gas.species_index(species) # [-] Species index
            EI_NOx                     += gas.Y[idx] * mdot_total / mdot_fuel # [kg/kg_fuel] Add contribution to NOx EI
        except ValueError:
            continue
    EI                                  = {
        'CO2': gas.Y[gas.species_index('CO2')] * mdot_total / mdot_fuel, # [kg/kg_fuel] CO2 emission index
        'CO': gas.Y[gas.species_index('CO')] * mdot_total / mdot_fuel,   # [kg/kg_fuel] CO emission index
        'H2O': gas.Y[gas.species_index('H2O')] * mdot_total / mdot_fuel, # [kg/kg_fuel] H2O emission index
        'NOx': EI_NOx,                                                   # [kg/kg_fuel] NOx emission index
    }
    return EI                                                            # [-] Return emission indices dictionary



# RCAIDE/Library/Methods/Powertrain/Converters/compressor/append_compressor_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_compressor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_compressor_conditions(compressor,segment,energy_conditions): 
    """
    Initializes empty condition containers for compressor analysis in the propulsion system.
    
    Parameters
    ----------
    compressor : Compressor
        The compressor component being analyzed
    segment : Segment
        The mission segment being analyzed
    energy_conditions : Conditions
        Container for storing energy system conditions
    
    Returns
    -------
    None
    
    Notes
    -----
    This function creates empty Conditions containers that will be populated
    during compressor performance calculations with thermodynamic states
    and operating parameters.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Compressor.compute_compressor_performance
    """
    
    ones_row    = segment.state.ones_row 
    energy_conditions.converters[compressor.tag]                                   = Conditions()
    energy_conditions.converters[compressor.tag].inputs                            = Conditions()
    energy_conditions.converters[compressor.tag].outputs                           = Conditions()
    energy_conditions.converters[compressor.tag].outputs.external_shaft_work_done  = 0*ones_row(1)
    energy_conditions.converters[compressor.tag].outputs.external_electrical_power = 0*ones_row(1)
    return 


# RCAIDE/Library/Methods/Powertrain/Converters/Compressor/__init__.py

"""
Collection of methods for computing compressor performance and conditions in propulsion systems. 
This module provides functionality for calculating compressor performance metrics and appending 
relevant thermodynamic conditions to the analysis.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters.Turbine
RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
        
from .compute_compressor_performance import compute_compressor_performance
from .append_compressor_conditions   import append_compressor_conditions


# RCAIDE/Library/Methods/Powertrain/Converters/Combustor/__init__.py

"""
Methods for modeling and analyzing combustors in propulsion systems.

This module provides functions for computing the thermodynamic performance of combustors
in gas turbine engines and other propulsion systems. It includes methods for calculating
combustion properties, fuel consumption, and temperature rise across the combustor, as well
as functions for initializing combustor conditions during mission analysis.

The combustor is a critical component in gas turbine engines where fuel is mixed with
compressed air and burned to produce high-temperature gases that drive the turbine.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters.Turbine
RCAIDE.Library.Methods.Powertrain.Converters.Compressor
RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle
""" 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
        
from .compute_combustor_performance         import compute_combustor_performance  
from .append_combustor_conditions           import append_combustor_conditions 




# RCAIDE/Methods/Energy/Propulsors/Turbojet/design_turbojet.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports     
import RCAIDE 
from RCAIDE.Framework.Mission.Common                                 import Conditions
from RCAIDE.Library.Methods.Powertrain.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Supersonic_Nozzle  import compute_supersonic_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet           import size_core  
from RCAIDE.Library.Methods.Powertrain                               import setup_operating_conditions 

# Python package imports   
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Turbojet
# ----------------------------------------------------------------------------------------------------------------------   
def design_turbojet(turbojet):  
    """
    Designs a turbojet engine by computing performance properties and sizing components based on design conditions.
    
    Parameters
    ----------
    turbojet : Turbojet
        Turbojet engine object containing design parameters and components
            - design_mach_number : float
                Design point Mach number [-]
            - design_altitude : float
                Design point altitude [m]
            - design_isa_deviation : float
                ISA temperature deviation [K]
            - working_fluid : Gas
                Working fluid object for gas properties
            - Components:
                - ram : Ram
                - inlet_nozzle : Compression_Nozzle
                - low_pressure_compressor : Compressor
                - high_pressure_compressor : Compressor
                - combustor : Combustor
                - high_pressure_turbine : Turbine
                - low_pressure_turbine : Turbine
                - core_nozzle : Supersonic_Nozzle
    
    Returns
    -------
    None
        Updates turbojet object attributes in-place:
            - mass_flow_rate_design : float
                Design core mass flow rate [kg/s]
            - design_core_massflow : float
                Core mass flow at design point [kg/s]
    
    Notes
    -----
    This function performs the following steps:
        1. Computes atmospheric conditions at design point
        2. Sets up freestream conditions
        3. Links and analyzes flow through each component:
            - Ram inlet
            - Inlet nozzle
            - Low pressure compressor
            - High pressure compressor
            - Combustor
            - High pressure turbine
            - Low pressure turbine
            - Core nozzle
        4. Sizes the core based on design thrust requirements
        5. Computes static sea level performance
    
    **Major Assumptions**
        * Quasi-one-dimensional flow
        * Each component operates in steady state
        * Perfect gas behavior in non-combustion sections
        * US Standard Atmosphere 1976 model
        * Earth gravity model
        * Design point defines core sizing
    
    **Theory**
    The design process follows standard gas turbine design principles:
    
    .. math::
        \\text{Mass flow continuity: } \\dot{m}_{in} = \\dot{m}_{out}
    
        \\text{Power Balance: } W_{compressor} = W_{turbine}
    
        \\text{Core sizing: } \\dot{m}_{core} = \\frac{F_{design}}{F_{sp} a_0}
    
    where:
        - :math:`F_{design}` is the design thrust
        - :math:`F_{sp}` is the specific thrust
        - :math:`a_0` is the freestream speed of sound
    
    References
    ----------
    [1] Mattingly, J. D., "Elements of Gas Turbine Propulsion", McGraw-Hill, 1996
    [2] Walsh, P. P., Fletcher, P., "Gas Turbine Performance", Blackwell Science, 2004
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet.compute_turbojet_performance
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbojet.size_core
    RCAIDE.Library.Methods.Powertrain.Propulsors.Common.compute_static_sea_level_performance
    """
    #check if mach number and temperature are passed
    if(turbojet.design_mach_number==None or turbojet.design_altitude==None):
        
        #raise an error
        raise NameError('The sizing conditions require an altitude and a Mach number')
    
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(turbojet.design_altitude,turbojet.design_isa_deviation)
        planet     = RCAIDE.Library.Attributes.Planets.Earth()
        
        p   = atmo_data.pressure          
        T   = atmo_data.temperature       
        rho = atmo_data.density          
        a   = atmo_data.speed_of_sound    
        mu  = atmo_data.dynamic_viscosity   
    
        # setup conditions
        conditions = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turbojet.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turbojet.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turbojet.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turbojet.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turbojet.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turbojet.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(a*turbojet.design_mach_number)
   
    segment                                        = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions                       = conditions 
    turbojet.append_operating_conditions(segment,conditions.energy,conditions.noise)        
    
    ram                       = turbojet.ram
    inlet_nozzle              = turbojet.inlet_nozzle
    low_pressure_compressor   = turbojet.low_pressure_compressor
    high_pressure_compressor  = turbojet.high_pressure_compressor
    combustor                 = turbojet.combustor
    high_pressure_turbine     = turbojet.high_pressure_turbine
    low_pressure_turbine      = turbojet.low_pressure_turbine
    core_nozzle               = turbojet.core_nozzle

    # unpack component conditions
    turbojet_conditions     = conditions.energy.propulsors[turbojet.tag]
    ram_conditions          = conditions.energy.converters[ram.tag]     
    inlet_nozzle_conditions = conditions.energy.converters[inlet_nozzle.tag]
    core_nozzle_conditions  = conditions.energy.converters[core_nozzle.tag] 
    lpc_conditions          = conditions.energy.converters[low_pressure_compressor.tag]
    hpc_conditions          = conditions.energy.converters[high_pressure_compressor.tag]
    lpt_conditions          = conditions.energy.converters[low_pressure_turbine.tag]
    hpt_conditions          = conditions.energy.converters[high_pressure_turbine.tag]
    combustor_conditions    = conditions.energy.converters[combustor.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turbojet.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle
    lpc_conditions.inputs.stagnation_temperature  = inlet_nozzle_conditions.outputs.stagnation_temperature
    lpc_conditions.inputs.stagnation_pressure     = inlet_nozzle_conditions.outputs.stagnation_pressure
    lpc_conditions.inputs.static_temperature      = inlet_nozzle_conditions.outputs.static_temperature
    lpc_conditions.inputs.static_pressure         = inlet_nozzle_conditions.outputs.static_pressure
    lpc_conditions.inputs.mach_number             = inlet_nozzle_conditions.outputs.mach_number  
    low_pressure_compressor.working_fluid         = inlet_nozzle.working_fluid
    low_pressure_compressor.reference_temperature = turbojet.reference_temperature
    low_pressure_compressor.reference_pressure    = turbojet.reference_pressure

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(low_pressure_compressor,conditions)

    # Step 7: Link the high pressure compressor to the low pressure compressor
    hpc_conditions.inputs.stagnation_temperature   = lpc_conditions.outputs.stagnation_temperature
    hpc_conditions.inputs.stagnation_pressure      = lpc_conditions.outputs.stagnation_pressure
    hpc_conditions.inputs.static_temperature       = lpc_conditions.outputs.static_temperature
    hpc_conditions.inputs.static_pressure          = lpc_conditions.outputs.static_pressure
    hpc_conditions.inputs.mach_number              = lpc_conditions.outputs.mach_number  
    high_pressure_compressor.working_fluid         = low_pressure_compressor.working_fluid  
    high_pressure_compressor.reference_temperature = turbojet.reference_temperature
    high_pressure_compressor.reference_pressure    = turbojet.reference_pressure   

    # Step 8: Compute flow through the high pressure compressor
    compute_compressor_performance(high_pressure_compressor,conditions)
   
    # Step 9: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature                = hpc_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = hpc_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = hpc_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = hpc_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = hpc_conditions.outputs.mach_number  
    combustor.working_fluid                                           = high_pressure_compressor.working_fluid  
    
    # Step 10: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,conditions)

    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number       
    hpt_conditions.inputs.compressor                = hpc_conditions.outputs   
    hpt_conditions.inputs.bypass_ratio              = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid 

    # Step 11: Compute flow through the high pressure turbine
    compute_turbine_performance(high_pressure_turbine,conditions)
            
    # Step 12: Link the low pressure turbine to the high pressure turbine
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number  
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid    
    lpt_conditions.inputs.compressor                 = lpc_conditions.outputs 
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.bypass_ratio               = 0.0 

    # Step 16: Compute flow through the low pressure turbine
    compute_turbine_performance(low_pressure_turbine,conditions)

    # Step 17: Link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid 

    # Step 18: Compute flow through the core nozzle
    compute_supersonic_nozzle_performance(core_nozzle,conditions)
 
    # Step 19: link the thrust component to the core nozzle
    turbojet_conditions.core_nozzle_area_ratio                   = core_nozzle_conditions.outputs.area_ratio 
    turbojet_conditions.core_nozzle_static_pressure              = core_nozzle_conditions.outputs.static_pressure
    turbojet_conditions.core_nozzle_exit_velocity                = core_nozzle_conditions.outputs.velocity 
    turbojet_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio 
    turbojet_conditions.stag_temp_lpt_exit                       = lpc_conditions.outputs.stagnation_temperature
    turbojet_conditions.stag_press_lpt_exit                      = lpc_conditions.outputs.stagnation_pressure 
    turbojet_conditions.total_temperature_reference              = lpc_conditions.outputs.stagnation_temperature
    turbojet_conditions.total_pressure_reference                 = lpc_conditions.outputs.stagnation_pressure 
    turbojet_conditions.flow_through_core                        = 1.0 #scaled constant to turn on core thrust computation
    turbojet_conditions.flow_through_fan                         = 0.0 #scaled constant to turn on fan thrust computation      
    
    # Step 20: Size the core of the turbojet  
    size_core(turbojet,conditions)
    
    # Step 21: Static Sea Level Thrust 
    atmo_data_sea_level   = atmosphere.compute_values(0.0,0.0)   
    V                     = atmo_data_sea_level.speed_of_sound[0][0]*0.01 
    operating_state       = setup_operating_conditions(turbojet, altitude = 0,velocity_range=np.array([V]))  
    operating_state.conditions.energy.propulsors[turbojet.tag].throttle[:,0] = 1.0  
    sls_T,_,sls_P,_,_,_                          = turbojet.compute_performance(operating_state) 
    turbojet.sealevel_static_thrust              = sls_T[0][0]
    turbojet.sealevel_static_power               = sls_P[0][0]
     
    return      
  


# RCAIDE/Library/Plots/Noise/__init__.py
# 

"""
RCAIDE Noise Plotting Package

This package contains modules for visualizing noise-related data and analysis 
results in RCAIDE.

See Also
--------
RCAIDE.Library.Plots : Parent plotting package
RCAIDE.Library.Methods.Noise : Noise analysis tools
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .plot_noise_level         import plot_noise_level
from .plot_3D_noise_contour    import plot_3D_noise_contour 
from .plot_2D_noise_contour    import plot_2D_noise_contour
from .post_process_noise_data  import post_process_noise_data


# RCAIDE/Methods/Powertrain/Sources/Batteries/__init__.py
# 

"""
This module provides functionality for modeling battery systems in powertrains. It includes methods for 
different battery chemistries such as Lithium-Ion LFP, Lithium-Ion NMC, and Aluminum-Air batteries,
as well as common battery functions.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Sources
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from . import Common
from . import Lithium_Ion_LFP
from . import Lithium_Ion_NMC
from . import Aluminum_Air


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_operating_items_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Operating Items Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_items_weight(vehicle):
    """
    Computes the weight of operating items using NASA FLOPS weight estimation method. 
    Includes crew, baggage, unusable fuel, engine oil, passenger service items, weapons, and cargo containers.

    Parameters
    ----------
    vehicle : Vehicle
        The vehicle instance containing:
            - networks : list
                Propulsion systems with:
                    - propulsors : list
                        Engine data for thrust and count
                    - fuel_lines : list
                        Fuel tank data
            - reference_area : float
                Wing reference area [m²]
            - mass_properties.max_zero_fuel : float
                Maximum zero fuel weight [kg]
            - flight_envelope.design_range : float
                Design range [nmi]
            - flight_envelope.design_mach_number : float
                Design cruise Mach number
            - mass_properties.cargo : float
                Cargo weight [kg]
            - passengers : int
                Total passenger count
            - fuselages : list
                Fuselage data with optional:
                    - number_coach_seats : int
                        Number of economy seats

    Returns
    -------
    output : Data
        Container with weight breakdown:
            - misc : float
                Weight of unusable fuel, engine oil, passenger service, cargo containers [kg]
            - flight_crew : float
                Flight crew and baggage weight [kg]
            - flight_attendants : float
                Flight attendant and galley crew weight [kg]
            - total : float
                Total operating items weight [kg]

    Notes
    -----
    Uses FLOPS correlations developed from commercial transport aircraft data. For more details, 
    please refer to the FLOPS documentation: https://ntrs.nasa.gov/citations/20170005851  

    **Major Assumptions**
        * If no tanks specified, assumes 5 fuel tanks
        * Default seat class distribution if not specified:
            - 5% first class
            - 10% business class
            - 85% economy class
        * If coach seats specified:
            - Remaining seats split 1/4 first class, 3/4 business
        * Crew requirements based on passenger count:
            - 2-3 flight crew (>150 pax = 3)
            - 1 attendant per 40 pax (min 1)
            - Additional galley crew for >250 pax

    **Theory**
    Component weights computed using empirical correlations:
    
    Unusable fuel weight:
    .. math::
        W_{UF} = 11.5N_{eng}T_{SLS}^{0.2} + 0.07S_{ref} + 1.6N_{tank}W_{ZF}^{0.28}

    Engine oil weight:
    .. math::
        W_{oil} = 0.082N_{eng}T_{SLS}^{0.65}

    Passenger service weight:
    .. math::
        W_{srv} = (5.164N_{1st} + 3.846N_{bus} + 2.529N_{eco})(R/M)^{0.255}
    
    Cargo container weight:
    .. math::
        W_{con} = 175 \\lceil \\frac{W_{cargo}}{950} \\rceil

    where:
        * N_eng = number of engines
        * T_SLS = sea level static thrust
        * S_ref = wing reference area
        * N_tank = number of fuel tanks
        * W_ZF = zero fuel weight
        * N_1st/bus/eco = number of seats by class
        * R = design range
        * M = design Mach number
        * W_{cargo} = cargo weight
    References
    ----------
    [1] NASA Flight Optimization System (FLOPS)
    """ 
    NENG =  0 
    for network in  vehicle.networks:
        for propulsor in network.propulsors: 
            ref_propulsor = propulsor  
            NENG  += 1
            
    THRUST          = ref_propulsor.sealevel_static_thrust * 1 / Units.lbf
    SW              = vehicle.reference_area / Units.ft ** 2
    FMXTOT          = vehicle.mass_properties.max_zero_fuel / Units.lbs
    DESRNG          = vehicle.flight_envelope.design_range / Units.nmi
    VMAX            = vehicle.flight_envelope.design_mach_number   
    
    number_of_tanks = 0  
    for network in  vehicle.networks:
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                number_of_tanks += 1 
    if number_of_tanks == 0:
        number_of_tanks = 5    
    
    WUF             = 11.5 * NENG * THRUST ** 0.2 + 0.07 * SW + 1.6 * number_of_tanks * FMXTOT ** 0.28  # unusable fuel weight
    WOIL            = 0.082 * NENG * THRUST ** 0.65  # engine oil weight
    
    for fuselage in  vehicle.fuselages: 
        if len(fuselage.cabins) > 0:
            NPT =  0
            NPF =  0
            NPB =  0
            for cabin in fuselage.cabins:
                for cabin_class in cabin.classes:
                    if type(cabin_class) == RCAIDE.Library.Components.Fuselages.Cabins.Classes.Economy:
                        NPT =  cabin_class.number_of_seats_abrest *  cabin_class.number_of_rows
                    elif type(cabin_class) == RCAIDE.Library.Components.Fuselages.Cabins.Classes.Business:
                        NPB =  cabin_class.number_of_seats_abrest *  cabin_class.number_of_rows
                    elif type(cabin_class) == RCAIDE.Library.Components.Fuselages.Cabins.Classes.First:
                        NPF =  cabin_class.number_of_seats_abrest *  cabin_class.number_of_rows
        else:
            NPF = vehicle.passengers / 20.
            NPB = vehicle.passengers / 10.
            NPT = vehicle.passengers - NPF - NPB
    vehicle.NPF = NPF
    vehicle.NPB = NPB
    vehicle.NPT = NPT
    WSRV        = (5.164 * NPF + 3.846 * NPB + 2.529 * NPT) * (DESRNG / VMAX) ** 0.255  # passenger service weight
    WCON        = 175 * np.ceil(vehicle.mass_properties.cargo / Units.lbs * 1. / 950)  # cargo container weight

    if vehicle.passengers >= 150:
        NFLCR = 3  # number of flight crew
        NGALC = 1 + np.floor(vehicle.passengers / 250.)  # number of galley crew
    else:
        NFLCR = 2
        NGALC = 0
    if vehicle.passengers < 51:
        NFLA = 1  # number of flight attendants, NSTU in FLOPS
    else:
        NFLA = 1 + np.floor(vehicle.passengers / 40.)

    WFLAAB = NFLA * 155 + NGALC * 200  # flight attendant weight, WSTUAB in FLOPS
    WFLCRB = NFLCR * 225  # flight crew and baggage weight

    output                           = Data()
    output.misc = WUF * Units.lbs + WOIL * Units.lbs + WSRV * Units.lbs + WCON * Units.lbs
    output.flight_crew               = WFLCRB * Units.lbs
    output.flight_attendants         = WFLAAB * Units.lbs
    output.total                     = output.misc + output.flight_crew + \
                                       output.flight_attendants
    return output


# RCAIDE/Library/Methods/Powertrain/Converters/Compressor/compute_compressor_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
# Imports 
# ---------------------------------------------------------------------------------------------------------------------- 
import numpy as np
from copy import  deepcopy

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_compression_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------
def compute_compressor_performance(compressor, conditions):
    """
    Computes the performance of a compressor based on its polytropic efficiency.

    Parameters
    ----------
    compressor : RCAIDE.Library.Components.Converters.Compressor
        Compressor component with the following attributes:
            - tag : str
                Identifier for the compressor
            - pressure_ratio : float
                Pressure ratio across compressor
            - polytropic_efficiency : float
                Polytropic efficiency of compression
            - working_fluid : Data
                Working fluid properties object
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - energy.converters[compressor.tag].inputs : Data
                Input conditions
                    - stagnation_temperature : numpy.ndarray
                        Inlet stagnation temperature [K]
                    - stagnation_pressure : numpy.ndarray
                        Inlet stagnation pressure [Pa]
                    - static_temperature : numpy.ndarray
                        Inlet static temperature [K]
                    - static_pressure : numpy.ndarray
                        Inlet static pressure [Pa]
                    - mach_number : numpy.ndarray
                        Inlet Mach number
            - energy.hybrid_power_split_ratio : float
                Ratio of power split for hybrid systems

    Returns
    -------
    None
        Results are stored in conditions.energy.converters[compressor.tag].outputs:
            - work_done : numpy.ndarray
                Specific work done by compressor [J/kg]
            - stagnation_temperature : numpy.ndarray
                Exit stagnation temperature [K]
            - stagnation_pressure : numpy.ndarray
                Exit stagnation pressure [Pa]
            - stagnation_enthalpy : numpy.ndarray
                Exit stagnation enthalpy [J/kg]
            - static_temperature : numpy.ndarray
                Exit static temperature [K]
            - static_pressure : numpy.ndarray
                Exit static pressure [Pa]
            - mach_number : numpy.ndarray
                Exit Mach number
            - gas_constant : numpy.ndarray
                Gas constant [J/(kg·K)]
            - gamma : numpy.ndarray
                Ratio of specific heats
            - cp : numpy.ndarray
                Specific heat at constant pressure [J/(kg·K)]

    Notes
    -----
    This function implements the thermodynamic calculations for a compressor with
    a specified pressure ratio and polytropic efficiency. The work done is adjusted
    by the hybrid power split ratio if applicable.
    
    **Major Assumptions**
        * Constant polytropic efficiency
        * Constant pressure ratio
        * Ideal gas behavior
        * Adiabatic process

    **Theory**
    The compression process follows the polytropic relation:

    .. math::
        T_{t,out}/T_{t,in} = (P_{t,out}/P_{t,in})^{(\\gamma-1)/(\\gamma \\eta_{p})}

    where :math:`\\eta_{p}` is the polytropic efficiency.

    Enthalpy is calculated using the specific heat at constant pressure and the stagnation temperature.

    .. math::
        h_{t} = C_{p} T_{t}

    References
    ----------
    [1] Cantwell, B., "AA283 Course Notes", Stanford University https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_BOOK/AA283_Aircraft_and_Rocket_Propulsion_BOOK_Brian_J_Cantwell_May_28_2024.pdf

    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Compressor.append_compressor_conditions
    """          
    
    # Unpack component inputs
    PR                    = compressor.pressure_ratio
    etapold               = compressor.polytropic_efficiency
    working_fluid         = compressor.working_fluid
    compressor_conditions = conditions.energy.converters[compressor.tag]
    Tt_in                 = compressor_conditions.inputs.stagnation_temperature
    Pt_in                 = compressor_conditions.inputs.stagnation_pressure 
    T0                    = compressor_conditions.inputs.static_temperature
    P0                    = compressor_conditions.inputs.static_pressure  
    M0                    = compressor_conditions.inputs.mach_number 
     
    # Compute the working fluid properties 
    gamma  = working_fluid.compute_gamma(T0,P0) 
    Cp     = working_fluid.compute_cp(T0,P0)    
    R      = working_fluid.compute_R(T0,P0)
        
    # Compute the output properties based on the pressure ratio of the component
    ht_in     = Tt_in*Cp 
    Pt_out    = Pt_in*PR
    Tt_out    = Tt_in*(PR**((gamma-1)/(gamma*etapold)))
    ht_out    = Tt_out*Cp
    T_out     = Tt_out/(1.+(gamma-1.)/2.*M0*M0)
    P_out     = Pt_out/((1.+(gamma-1.)/2.*M0*M0)**(gamma/(gamma-1.))) 
    M_out     = np.sqrt( (((Pt_out/P_out)**((gamma-1.)/gamma))-1.) *2./(gamma-1.) )
    
    # Compute the work done by the compressor (normalized by mass flow i.e. J/(kg/s)
    work_done = ht_out - ht_in
    
    phi       =  conditions.energy.hybrid_power_split_ratio 

    # Pack results  
    compressor_conditions.outputs.work_done               = (1-phi)*work_done
    compressor_conditions.outputs.stagnation_temperature  = Tt_out
    compressor_conditions.outputs.stagnation_pressure     = Pt_out
    compressor_conditions.outputs.stagnation_enthalpy     = ht_out
    compressor_conditions.outputs.static_temperature      = T_out
    compressor_conditions.outputs.static_pressure         = P_out 
    compressor_conditions.outputs.mach_number             = M_out
    compressor_conditions.outputs.gas_constant            = R
    compressor_conditions.outputs.gamma                   = gamma 
    compressor_conditions.outputs.cp                      = Cp
    
    return 




# RCAIDE/Methods/Energy/Distributors/Electrical_Bus.py
# 
# 
# Created: Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# imports 
import numpy as np

 
# ----------------------------------------------------------------------------------------------------------------------
# compute_bus_conditions
# ----------------------------------------------------------------------------------------------------------------------
def compute_bus_conditions(bus, state, t_idx, delta_t): 
    """
    Computes the conditions of the bus based on the response of the battery modules.
    
    Parameters
    ----------
    bus : ElectricalBus
        The electrical bus component with the following attributes:
            - tag : str
                Identifier for the bus
            - battery_modules : list
                List of battery modules connected to the bus
            - battery_module_electric_configuration : str
                Configuration of battery modules ('Series' or 'Parallel')
    state : State
        Current system state containing conditions for all components
    t_idx : int
        Current time index in the simulation
    delta_t : float
        Time step [s]
    
    Returns
    -------
    None
    
    Notes
    -----
    This function calculates the electrical properties of the bus based on the connected
    battery modules. It handles both series and parallel configurations of battery modules.
    
    For series configuration:
        - Voltage is the sum of all battery module voltages
        - Temperature and energy are averaged/summed across modules
        - State of charge is taken from the last battery module
    
    For parallel configuration:
        - Voltage is the same as the last battery module
        - Temperature and energy are averaged/summed across modules
        - State of charge is taken from the last battery module's cell
    
    The function also handles the fully charged state by setting charging current,
    power draw, and current draw to zero when the state of charge reaches 1.0.
    
    **Major Assumptions**
        * Battery modules are the primary energy source for the bus
        * In parallel configuration, all modules have the same voltage
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Distributors.Electrical_Bus.append_bus_conditions
    RCAIDE.Library.Methods.Powertrain.Energy_Storage.Battery.compute_battery_module_conditions
    """
    bus_conditions = state.conditions.energy[bus.tag]
    phi   = state.conditions.energy.hybrid_power_split_ratio
     
    if len(bus.battery_modules) != 0: 
        if bus.battery_module_electric_configuration == 'Series':
            bm_conditions                               = [bus_conditions.battery_modules[bm.tag] for bm in bus.battery_modules]
            bus_conditions.voltage_open_circuit[t_idx]  = sum(bm.voltage_open_circuit[t_idx] for bm in bm_conditions)
            bus_conditions.voltage_under_load[t_idx]    = sum(bm.voltage_under_load[t_idx] for bm in bm_conditions)
            bus_conditions.heat_energy_generated[t_idx] = sum(bm.heat_energy_generated[t_idx] for bm in bm_conditions)
            bus_conditions.efficiency[t_idx]            = (bus_conditions.power_draw[t_idx]*phi[t_idx] + bus_conditions.heat_energy_generated[t_idx])/(bus_conditions.power_draw[t_idx]*phi[t_idx])
            if t_idx != state.numerics.number_of_control_points-1:  
                bm_conditions                              = [bus_conditions.battery_modules[bm.tag] for bm in bus.battery_modules]
                bus_conditions.temperature[t_idx+1]        = sum(bm.temperature[t_idx+1] for bm in bm_conditions)/ len(bus.battery_modules)
                bus_conditions.energy[t_idx+1]             = sum(bm.energy[t_idx+1] for bm in bm_conditions)
                bus_conditions.state_of_charge[t_idx+1]    = bm_conditions[-1].state_of_charge[t_idx+1]
    
        elif bus.battery_module_electric_configuration == 'Parallel':
            bm_conditions                               = [bus_conditions.battery_modules[bm.tag] for bm in bus.battery_modules]
            bus_conditions.heat_energy_generated[t_idx] = sum(bm.heat_energy_generated[t_idx] for bm in bm_conditions)
            bus_conditions.voltage_open_circuit[t_idx]  = bm_conditions[-1].voltage_open_circuit[t_idx]
            bus_conditions.voltage_under_load[t_idx]    = bm_conditions[-1].voltage_under_load[t_idx]             
            bus_conditions.efficiency[t_idx]            = (bus_conditions.power_draw[t_idx]*phi[t_idx] +  bus_conditions.heat_energy_generated[t_idx])/(bus_conditions.power_draw[t_idx]*phi[t_idx])
            if t_idx != state.numerics.number_of_control_points-1:  
                bus_conditions.heat_energy_generated[t_idx] = sum(bm.heat_energy_generated[t_idx] for bm in bm_conditions)
                bus_conditions.temperature[t_idx+1]         = sum(bm.temperature[t_idx+1] for bm in bm_conditions)/len(bus.battery_modules)
                bus_conditions.energy[t_idx+1]              = sum(bm.energy[t_idx+1] for bm in bm_conditions)
                bus_conditions.state_of_charge[t_idx+1]     = bm_conditions[-1].cell.state_of_charge[t_idx+1]
        
    if t_idx != state.numerics.number_of_control_points-1:  
        # Handle fully charged state
        if state.conditions.energy.recharging and np.float16(bus_conditions.state_of_charge[t_idx+1]) == 1:
            bus_conditions.charging_current[t_idx+1] = 0
            bus_conditions.power_draw[t_idx+1]       = 0
            bus_conditions.current_draw[t_idx+1]     = 0
    return


#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core                              import Units
from RCAIDE.Library.Methods.Powertrain                  import setup_operating_conditions  

# ------------------------------------------------------------------------------
#   Rotor Analysis
# ------------------------------------------------------------------------------ 
def rotor_aerodynamic_analysis(rotor,
                           velocity_range,
                           angular_velocity = 2500*Units.rpm,
                           blade_pitch_command = 0, 
                           angle_of_attack = 0, 
                           altitude = 0,
                           design_flag=False):
       
    operating_state = setup_operating_conditions(rotor, altitude = altitude,velocity_range=velocity_range, angle_of_attack=angle_of_attack)
         
    rotor_conditions                             = operating_state.conditions.energy.converters[rotor.tag]
    rotor_conditions.design_flag                 = design_flag
    rotor_conditions.omega[:,0]                  = angular_velocity
    rotor_conditions.blade_pitch_command[:,0]    = blade_pitch_command  
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance(rotor,operating_state.conditions)
     
    results = operating_state.conditions.energy.converters[rotor.tag] 
    return  results


## @defgroup Library-Plots-Energy 
# RCAIDE/Library/Plots/Energy/__init__.py
# 

"""
RCAIDE Emissions Plotting Package

This module provides visualization tools for analyzing and displaying emissions-related 
data from vehicle and mission simulations. It focuses on greenhouse gas emissions.
"""
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .plot_CO2e_emissions      import  plot_CO2e_emissions


# RCAIDE/Library/Methods/Powertrain/Propulsors/Internal_Combustion_Engine/append_internal_combustion_engine_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions 

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_internal_combustion_engine_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_internal_combustion_engine_conditions(propulsor, segment, energy_conditions, noise_conditions):
    """
    Initializes internal combustion engine operating conditions for a mission segment.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Internal_Combustion_Engine
        Internal combustion engine propulsor component with the following attributes:
            - tag : str
                Identifier for the engine
            - items : dict
                Dictionary of subcomponents
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where engine conditions will be stored
    noise_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Noise conditions container where engine noise conditions will be stored
    
    Returns
    -------
    None
        Results are stored in energy_conditions.propulsors[propulsor.tag] and
        segment.state.conditions.noise[propulsor.tag]
    
    Notes
    -----
    This function initializes the necessary data structures for storing internal combustion
    engine operating conditions during a mission segment. It creates zero-filled arrays
    for various performance parameters and recursively calls the append_operating_conditions
    method for each subcomponent of the engine.
    
    The function initializes the following parameters in energy_conditions:
        * throttle
        * commanded_thrust_vector_angle
        * thrust
        * power
        * moment
        * fuel_flow_rate
        * inputs and outputs containers
    
    It also creates a noise conditions container for the engine.
    
    **Major Assumptions**
        * All arrays are initialized with zeros
        * Each component has an append_operating_conditions method
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Internal_Combustion_Engine.compute_internal_combustion_engine_performance
    """
    # unpack 
    ones_row          = segment.state.ones_row 
    
    # add propulsor conditions 
    energy_conditions.propulsors[propulsor.tag]                               = Conditions()  
    energy_conditions.propulsors[propulsor.tag].throttle                      = 0. * ones_row(1)      
    energy_conditions.propulsors[propulsor.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    energy_conditions.propulsors[propulsor.tag].thrust                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].power                         = 0. * ones_row(1) 
    energy_conditions.propulsors[propulsor.tag].moment                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].fuel_flow_rate                = 0. * ones_row(1)
    energy_conditions.propulsors[propulsor.tag].inputs                        = Conditions()
    energy_conditions.propulsors[propulsor.tag].outputs                       = Conditions() 
    segment.state.conditions.noise[propulsor.tag]                             = Conditions()  
 
    for tag, item in  propulsor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,energy_conditions,noise_conditions=noise_conditions) 
            for sub_tag, sub_item in  item.items(): 
                if issubclass(type(sub_item), RCAIDE.Library.Components.Component): 
                    sub_item.append_operating_conditions(segment,energy_conditions)                     
    return 


# RCAIDE/Methods/Energy/Propulsion/Converters/Common/__init__.py
# 

"""
This module provides functionality for setting up and managing systems that draw power from the powertrain system, such as avionics and payloads. 
It includes methods for configuring operating conditions and appending avionics and payload conditions to simulation results.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters
RCAIDE.Library.Methods.Powertrain.Sources
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_avionics_conditions             import append_avionics_conditions
from .append_payload_conditions              import append_payload_conditions 


# RCAIDE/Framework/Analyses/Mission/Segments/Untrimmed/__init__.py
# 

"""
Mission segemtn for analyzing aircraft performance in untrimmed flight conditions, where forces and moments may not be in equilibrium.

See Also
--------
RCAIDE.Library.Mission.Segments.Cruise
RCAIDE.Library.Mission.Segments.Transition
RCAIDE.Library.Mission.Segments.Climb
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Untrimmed 


# RCAIDE/Library/Methods/Powertrain/Converters/Turboelectric_Generator/append_turboelectric_generator_conditions.py 
# 
# Created:  Feb 2025, M. Clarke  
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_turboelectric_generator_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turboelectric_generator_conditions(turboelectric_generator,segment,energy_conditions):  
    """
    Initializes and appends operating conditions data structures for a turboelectric generator to the energy conditions structure.
    
    Parameters
    ----------
    turboelectric_generator : RCAIDE.Components.Energy.Converters.Turboelectric_Generator
        The turboelectric generator component for which conditions are being appended
    segment : RCAIDE.Analyses.Mission.Segments
        The mission segment being evaluated
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Container for energy-related conditions during the mission segment
        
    Returns
    -------
    None
        This function modifies the energy_conditions object in-place
        
    Notes
    -----
    This function initializes the condition structure for a turboelectric generator
    and its subcomponents (turboshaft and generator) with zero values, then calls
    the respective append_operating_conditions methods for each subcomponent.
    """

    ones_row    = segment.state.ones_row   
 
    energy_conditions.converters[turboelectric_generator.tag] = Conditions() 
    energy_conditions.converters[turboelectric_generator.tag].throttle                                   = 0. * ones_row(1)     
    energy_conditions.converters[turboelectric_generator.tag].commanded_thrust_vector_angle              = 0. * ones_row(1)   
    energy_conditions.converters[turboelectric_generator.tag].power                                      = 0. * ones_row(1)
    energy_conditions.converters[turboelectric_generator.tag].fuel_flow_rate                             = 0. * ones_row(1)
    energy_conditions.converters[turboelectric_generator.tag].inputs                                     = Conditions()
    energy_conditions.converters[turboelectric_generator.tag].outputs                                    = Conditions() 
  
    turboshaft = turboelectric_generator.turboshaft
    generator  = turboelectric_generator.generator
    turboshaft.append_operating_conditions(segment,energy_conditions)
    generator.append_operating_conditions(segment,energy_conditions)
    return 


# RCAIDE/Library/Attributes/Planets/Earth.py
# 
#
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------  
#  Imports
# ----------------------------------------------------------------------------------------------------------------------  
 
from .Planet                import Planet 

# ----------------------------------------------------------------------------------------------------------------------  
#  Earth Class
# ----------------------------------------------------------------------------------------------------------------------  
class Earth(Planet):
    """
    A class representing Earth's physical properties and gravitational characteristics.

    Attributes
    ----------
    tag : str
        Identifier for the planet ('Earth')
    mass : float
        Mass of Earth in kg (5.98e24)
    mean_radius : float
        Average radius of Earth in m (6.371e6)
    sea_level_gravity : float
        Gravitational acceleration at sea level in m/s² (9.80665)

    Notes
    -----
    This class implements standard Earth properties used in aerospace calculations.
    Values are based on internationally accepted standards for Earth's physical parameters.
    """
    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """         
        self.tag = 'Earth'
        self.mass              = 5.98e24  # kg
        self.mean_radius       = 6.371e6  # m
        self.sea_level_gravity = 9.80665  # m/s^2    

    def compute_gravity(self, H=0.0):
        """
        Compute the gravitational acceleration at a given altitude above Earth's surface.

        Parameters
        ----------
        H : float, optional
            Altitude above sea level in meters. Default is 0.0

        Returns
        -------
        gh : float
            Gravitational acceleration at the specified altitude in m/s²

        Notes
        -----
        **Theory**
        Uses the inverse square law for gravitational acceleration:
        .. math::
            g(h) = g_0 \\left(\\frac{R_e}{R_e + h}\\right)^2

        where:
            - g₀ is sea level gravity
            - Rₑ is Earth's mean radius
            - h is altitude above sea level
        """          
        # Unpack
        g0  = self.sea_level_gravity
        Re  = self.mean_radius 
        
        # Calculate gravity
        gh = g0*(Re/(Re+H))**2.0
        
        return gh


# RCAIDE/Library/Attributes/Solids/Aluminum.py
# 
 

# Created: Mar 2024 M. Clarke

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Units
from .Solid import Solid 

#-------------------------------------------------------------------------------
# Aluminum for WavyChannel for battery cooling
#------------------------------------------------------------------------------- 
class Aluminum(Solid): 
    """ 
    A class representing 6061-T6 aluminum and its material properties.

    Attributes
    ----------
    density : float
        Material density in kg/m³ (2700)
    thermal_conductivity : float
        Heat conduction coefficient in W/(m·K) (202.4)
    specific_heat_capacity : float
        Specific heat at constant pressure in J/(kg·K) (871)
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (310e6)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (206e6)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (607e6)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (276e6)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (206e6)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (386e6)
    minimum_gage_thickness : float
        Minimum manufacturable thickness in m (0.0)

    Notes
    -----
    This class implements standard 6061-T6 aluminum properties for both structural 
    and thermal applications. The thermal properties are particularly relevant for 
    battery cooling applications.

    **Definitions**
    
    'Ultimate Strength'
        The maximum stress that a material can withstand before failure
    
    'Yield Strength'
        The stress at which a material begins to deform plastically
    
    'Thermal Conductivity'
        The property of a material to conduct heat, measured in watts per meter-kelvin

    References
    ----------
    [1] MatWeb. (n.d.). Aluminum 6061-T6; 6061-T651. https://www.matweb.com/search/DataSheet.aspx?MatGUID=b8d536e0b9b54bd7b69e4124d8f1d20a 
    """

    def __defaults__(self):
        """Sets material properties at instantiation. 

        Assumptions:
            None
    
        Source:
            Cao W, Zhao C, Wang Y, et al. Thermal modeling of full-size-scale cylindrical battery pack cooled
            by channeled liquid flow[J]. International journal of heat and mass transfer, 2019, 138: 1178-1187. 
        """

        self.density                    = 2719
        self.thermal_conductivity       = 202.4
        self.specific_heat_capacity     = 871 
        self.ultimate_tensile_strength  = 310e6 * Units.Pa
        self.ultimate_shear_strength    = 206e6 * Units.Pa
        self.ultimate_bearing_strength  = 607e6 * Units.Pa
        self.yield_tensile_strength     = 276e6 * Units.Pa
        self.yield_shear_strength       = 206e6 * Units.Pa
        self.yield_bearing_strength     = 386e6 * Units.Pa
        self.minimum_gage_thickness     = 0.0   * Units.m
        self.density                    = 2700. * Units['kg/(m**3)']        



# RCAIDE/Library/Attributes/LH2.py
# 
# 
# Created:  Sep 2023, M. Clarke
# Modified: 
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .Propellant import Propellant
from RCAIDE.Framework.Core import Units

# ----------------------------------------------------------------------------------------------------------------------
#  Rocket_LH2
# ----------------------------------------------------------------------------------------------------------------------  
class Rocket_LH2(Propellant):
    """
    A class representing liquid hydrogen (LH2) fuel properties specifically for rocket 
    propulsion applications. Optimized for high-performance rocket engines with 
    oxygen as oxidizer.

    Attributes
    ----------
    tag : str
        Identifier for the propellant ('Rocket_LH2')
    molecular_weight : float
        Molecular weight in kg/kmol (12.644)
    isentropic_expansion_factor : float
        Ratio of specific heats (1.145)
    combustion_temperature : float
        Adiabatic flame temperature in K (3331.0)
    gas_specific_constant : float
        Specific gas constant in J/(kg*K) (8314.45986/molecular_weight)

    Notes
    -----
    This class implements properties for rocket-grade liquid hydrogen, focusing on 
    parameters relevant to high-performance rocket engine applications. Properties 
    are optimized for use with liquid oxygen in bipropellant rocket engines.

    **Definitions**
    
    'Isentropic Expansion Factor'
        Ratio of specific heats (cp/cv) for exhaust products
    
    'Combustion Temperature'
        Theoretical maximum temperature achieved in the combustion chamber
    
    'Gas Specific Constant'
        Individual gas constant for combustion products

    **Major Assumptions**
        * Properties are for rocket engine operating conditions
        * O/F ratio of 5.50 (oxygen to fuel mass ratio)

    References
    ----------
    [1] Sutton, G. P., & Biblarz, O. (2017). Rocket Propulsion Elements. John Wiley & Sons Inc. 
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Assujmes an O/F ratio 5.50 
        
        Source:
            Sutton, Rocket Propulsion Elements Using CEA
        """    
        self.tag                         = 'Rocket_LH2'
        self.molecular_weight            = 12.644                             # [kg/kmol]
        self.isentropic_expansion_factor = 1.145
        self.combustion_temperature      = 3331.0*Units.kelvin                # [K]                      
        self.gas_specific_constant       = (8314.45986/self.molecular_weight)*Units['J/(kg*K)'] # [J/(kg-K)]

        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''


# RCAIDE/Library/Methods/Powertrain/Converters/Fan/compute_fan_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# package imports
import numpy as np 
from RCAIDE.Library.Methods.Gas_Dynamics.fm_id import fm_id

# exceptions/warnings
from warnings import warn

# ----------------------------------------------------------------------------------------------------------------------
#  compute_expansion_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------        
def compute_expansion_nozzle_performance(expansion_nozzle, conditions):
    """
    Computes the thermodynamic performance of an expansion nozzle in a propulsion system.
    
    Parameters
    ----------
    expansion_nozzle : RCAIDE.Library.Components.Converters.Expansion_Nozzle
        Expansion nozzle component with the following attributes:
            - tag : str
                Identifier for the nozzle
            - working_fluid : Data
                Working fluid properties object
            - pressure_ratio : float
                Pressure ratio across the nozzle
            - polytropic_efficiency : float
                Polytropic efficiency of the expansion process
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - freestream : Data
                Freestream properties
                - isentropic_expansion_factor : numpy.ndarray
                    Ratio of specific heats (gamma)
                - specific_heat_at_constant_pressure : numpy.ndarray
                    Specific heat at constant pressure [J/(kg·K)]
                - pressure : numpy.ndarray
                    Freestream pressure [Pa]
                - stagnation_pressure : numpy.ndarray
                    Freestream stagnation pressure [Pa]
                - stagnation_temperature : numpy.ndarray
                    Freestream stagnation temperature [K]
                - specific_gas_constant : numpy.ndarray
                    Specific gas constant [J/(kg·K)]
                - mach_number : numpy.ndarray
                    Freestream Mach number
            - energy : Data
                Energy conditions
                    - converters : dict
                        Converter energy conditions indexed by tag
    
    Returns
    -------
    None
        Results are stored in conditions.energy.converters[expansion_nozzle.tag].outputs:
            - stagnation_temperature : numpy.ndarray
                Stagnation temperature at nozzle exit [K]
            - stagnation_pressure : numpy.ndarray
                Stagnation pressure at nozzle exit [Pa]
            - stagnation_enthalpy : numpy.ndarray
                Stagnation enthalpy at nozzle exit [J/kg]
            - mach_number : numpy.ndarray
                Mach number at nozzle exit
            - static_temperature : numpy.ndarray
                Static temperature at nozzle exit [K]
            - static_enthalpy : numpy.ndarray
                Static enthalpy at nozzle exit [J/kg]
            - velocity : numpy.ndarray
                Exit velocity [m/s]
            - static_pressure : numpy.ndarray
                Static pressure at nozzle exit [Pa]
            - area_ratio : numpy.ndarray
                Exit to freestream area ratio
    
    Notes
    -----
    This function computes the thermodynamic properties at the expansion nozzle exit based on
    the inlet conditions and nozzle characteristics. It calculates the exit velocity, pressure,
    temperature, and other properties for both subsonic and supersonic flow regimes.
    
    The computation follows these steps:
        1. Extract freestream and inlet conditions
        2. Compute working fluid properties (gamma, Cp)
        3. Calculate stagnation conditions at exit using pressure ratio
        4. Compute exit Mach number based on pressure ratio
        5. Handle subsonic (M < 1.0) and supersonic (M ≥ 1.0) cases separately
        6. Calculate static conditions (temperature, pressure) at exit
        7. Compute exit velocity from energy conservation
        8. Calculate area ratio between freestream and nozzle exit
        9. Store all results in the conditions data structure
    
    **Major Assumptions**
        * Constant polytropic efficiency and pressure ratio
        * If pressures make the Mach number go negative, these values are corrected
    
    **Theory**
    For subsonic flow (M < 1.0), the exit pressure equals the ambient pressure, and the
    Mach number is calculated from the pressure ratio. For supersonic flow (M ≥ 1.0),
    the nozzle is choked, and the exit pressure is calculated from the Mach number.
    
    The exit velocity is calculated from the conservation of energy:
    
    .. math::
        u_{out} = \\sqrt{2(h_{t,out} - h_{out})}
    
    where h_{t,out} is the exit stagnation enthalpy and h_{out} is the exit static enthalpy.
    
    References
    ----------
    [1] Cantwell, B., "AA283 Course Notes", Stanford University https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/
    
    See Also
    --------
    RCAIDE.Library.Methods.Gas_Dynamics.fm_id
    """                 
    # Unpack flight conditions     
    M0                = conditions.freestream.mach_number
    P0                = conditions.freestream.pressure
    Pt0               = conditions.freestream.stagnation_pressure
    Tt0               = conditions.freestream.stagnation_temperature
    nozzle_conditions = conditions.energy.converters[expansion_nozzle.tag]
    
    # Unpack exansion nozzle inputs
    Tt_in    = nozzle_conditions.inputs.stagnation_temperature
    Pt_in    = nozzle_conditions.inputs.stagnation_pressure 
    PR       = expansion_nozzle.pressure_ratio
    etapold  = expansion_nozzle.polytropic_efficiency
    
    P_in  = nozzle_conditions.inputs.static_pressure    
    T_in  = nozzle_conditions.inputs.static_temperature
    
    # Unpack ram inputs       
    working_fluid  = expansion_nozzle.working_fluid 
    gamma          = working_fluid.compute_gamma(T_in,P_in) 
    Cp             = working_fluid.compute_cp(T_in,P_in)    
     
    # Compute output stagnation quantities
    Pt_out   = Pt_in*PR
    Tt_out   = Tt_in*PR**((gamma-1)/(gamma)*etapold)
    ht_out   = Cp*Tt_out
    
    # A cap so pressure doesn't go negative
    Pt_out[Pt_out<P0] = P0[Pt_out<P0]
    
    # Compute the output Mach number, static quantities and the output velocity
    Mach          = np.sqrt((((Pt_out/P0)**((gamma-1)/gamma))-1)*2/(gamma-1)) 
    
    #initializing the Pout array
    P_out         = np.ones_like(Mach)
    
    # Computing output pressure and Mach number for the case Mach <1.0
    i_low         = Mach < 1.0
    P_out[i_low]  = P0[i_low]
    Mach[i_low]   = np.sqrt((((Pt_out[i_low]/P0[i_low])**((gamma[i_low]-1.)/gamma[i_low]))-1.)*2./(gamma[i_low]-1.))
    
    # Computing output pressure and Mach number for the case Mach >=1.0     
    i_high        = Mach >=1.0   
    Mach[i_high]  = Mach[i_high]/Mach[i_high]
    P_out[i_high] = Pt_out[i_high]/(1.+(gamma[i_high]-1.)/2.*Mach[i_high]*Mach[i_high])**(gamma[i_high]/(gamma[i_high]-1.))
    
    # A cap to make sure Mach doesn't go to zero:
    if np.any(Mach<=0.0):
        warn('Pressures Result in Negative Mach Number, making positive',RuntimeWarning)
        Mach[Mach<=0.0] = 0.001
    
    # Compute the output temperature,enthalpy,velocity and density
    T_out         = Tt_out/(1+(gamma-1)/2*Mach*Mach)
    h_out         = T_out * Cp
    u_out         = np.sqrt(2*(ht_out-h_out))
    #rho_out       = P_out/(R*T_out)
    
    # Compute the freestream to nozzle area ratio  
    area_ratio    = (fm_id(M0,gamma)/fm_id(Mach,gamma)*(1/(Pt_out/Pt0))*(np.sqrt(Tt_out/Tt0)))
    
    #pack computed quantities into outputs
    nozzle_conditions.outputs.area_ratio              = area_ratio
    nozzle_conditions.outputs.mach_number             = Mach
    #nozzle_conditions.outputs.density                 = rho_out
    nozzle_conditions.outputs.velocity                = u_out
    nozzle_conditions.outputs.static_pressure         = P_out
    nozzle_conditions.outputs.static_temperature      = T_out
    nozzle_conditions.outputs.static_enthalpy         = h_out
    nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    
    return 



# RCAIDE/Library/Compoments/Powertrain/Distributors/__init__.py
# 

"""
Energy Distributor module providing components for aircraft power distribution

This module contains implementations for various energy distributor components including
coolant lines, fuel lines, and electrical busses. These components serve
as the primary energy distributors in aircraft propulsion systems.

See Also
--------
RCAIDE.Library.Components.Powertrain.Modulators
    Energy control systems
RCAIDE.Library.Components.Powertrain.Sources
    Fuel storage and delivery systems
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Coolant_Line                         import Coolant_Line
from .Electrical_Bus                       import Electrical_Bus
from .Fuel_Line                            import Fuel_Line




# RCAIDE/Library/Methods/Weights/Correlation_Buildups/BWB/compute_cabin_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from RCAIDE.Framework.Core import Units

# ---------------------------------------------------------------------------------------------------------------------- 
#  Cabin Weight 
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_cabin_weight(cabin_area, TOGW):
    """
    Computes the structural weight of the pressurized cabin section of a BWB aircraft using 
    regression-based methods derived from FEA studies.

    Parameters
    ----------
    cabin_area : float
        Planform area of pressurized passenger cabin [m²]
    TOGW : float
        Takeoff gross weight of aircraft [kg]

    Returns
    -------
    W_cabin : float
        Estimated structural weight of BWB pressurized cabin section [kg]

    Notes
    -----
    Uses regression equations derived from detailed FEA studies by Bradley to estimate the structural
    weight required for the non-circular pressure vessel of a BWB cabin [1].

    **Major Assumptions**
        * Pressurized sandwich composite structure
        * Ultimate cabin pressure differential of 18.6 psi
        * Critical load case is 2.5g maneuver at maximum TOGW
        * Uniform pressure distribution
        * Non-buckling design
        * Structural efficiency similar to baseline FEA model

    **Theory**
    Weight is computed using:
    .. math::
        W_{cabin} = 5.698865 \\cdot 0.316422 \\cdot W^{0.166552} \\cdot S_{cab}^{1.061158}

    where:
        * W = takeoff weight
        * S_cab = cabin planform area

    The equation coefficients account for:
        * Non-circular pressure vessel penalties
        * Composite material properties
        * Combined pressure and flight loads
        * Minimum gauge requirements

    References
    ----------
    [1] Bradley, K. R., "A Sizing Methodology for the Conceptual Design of Blended-Wing-Body Transports," NASA/CR-2004-213016, 2004.
    See Also
    --------
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.BWB.FLOPS.compute_aft_centerbody_weight
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.BWB.FLOPS.compute_operating_empty_weight
    """       
    
    # convert to imperial units
    S_cab    = cabin_area / Units.feet ** 2.0
    W        = TOGW       / Units.pounds
    
    W_cabin = 5.698865 * 0.316422 * (W ** 0.166552) * S_cab ** 1.061158
    
    # convert to SI units
    W_cabin = W_cabin * Units.pounds
    
    return W_cabin


# RCAIDE/Library/Methods/Energy/Fuel_Cells/__init__.py
# 

"""
This module provides methods for modeling and analyzing fuel cell systems in aircraft powertrains. 

It includes functionality for different fuel cell types and modeling approaches. The module 
contains specialized submodules for common fuel cell operations, Larminie-Dicks model implementation, 
and Proton Exchange Membrane (PEM) fuel cell modeling. These methods support the design, performance 
analysis, and mission simulation of fuel cell-powered aircraft.

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.Generic_Fuel_Cell_Stack
RCAIDE.Library.Methods.Powertrain.Converters
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from . import Common
from . import Larminie_Model
from . import Proton_Exchange_Membrane


# RCAIDE/Library/Missions/Segments/Cruise/Constant_Mach_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE  
import RCAIDE  

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ---------------------------------------------------------------------------------------------------------------------- 
def initialize_conditions(segment):
    """
    Initializes conditions for constant Mach cruise at fixed altitude

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed        
            - altitude : float
                Cruise altitude [m]
            - distance : float
                Ground distance to cover [m]
            - mach_number : float
                Mach number to maintain [-]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
            - analyses:
                atmosphere : Model
                    Atmospheric model for property calculations

    Notes
    -----
    This function sets up the initial conditions for a cruise segment with constant
    Mach number and constant altitude. The true airspeed is determined from the
    Mach number and local speed of sound.

    **Calculation Process**
        1. Get atmospheric properties at altitude
        2. Calculate true airspeed from Mach number:
            V = M * a where:
                - M is Mach number
                - a is speed of sound
        3. Calculate time required based on distance and speed
        4. Discretize time points
        5. Decompose velocity into components using sideslip angle

    **Major Assumptions**
        * Constant Mach number
        * Constant altitude
        * Standard atmosphere model
        * Small angle approximations
        * Quasi-steady flight
        * No wind effects

    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.position_vector [m]
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.time [s]

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Update.atmosphere
    """      
    
    # unpack
    alt        = segment.altitude
    xf         = segment.distance
    mach       = segment.mach_number
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions    
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]    
    segment.state.conditions.freestream.altitude[:,0] = alt
    
    # Update freestream to get speed of sound
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment)
    a          = conditions.freestream.speed_of_sound       

    # check for initial velocity
    if mach is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])    
    else:# compute speed, constant with constant altitude
        air_speed = mach * a
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = xf / air_speed + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      =  t_nondim * (t_final-t_initial) + t_initial
    v_x       = np.cos(beta)*air_speed 
    v_y       = np.sin(beta)*air_speed 
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]
    



# RCAIDE/Library/Missions/Segments/Climb/Constant_Speed_Linear_Altitude.py 
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """
    Initializes conditions for constant speed climb with linear altitude change

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the initial conditions for a climb segment with constant
    true airspeed and linear altitude variation. The climb angle is determined by
    the distance and altitude change.

    **Required Segment Components**

    segment:
        - air_speed : float
            True airspeed to maintain [m/s]
        - altitude_start : float
            Initial altitude [m]
        - altitude_end : float
            Final altitude [m]
        - distance : float
            Ground distance to cover [m]
        - sideslip_angle : float
            Aircraft sideslip angle [rad]
        - state:
            numerics.dimensionless.control_points : array
                Discretization points [-]
            conditions : Data
                State conditions container

    **Calculation Process**
        1. Calculate climb angle from altitude change and distance
        2. Discretize altitude profile
        3. Decompose constant velocity into components using:
            - Computed climb angle
            - Sideslip angle
            - Constant speed requirement

    **Major Assumptions**
        * Constant true airspeed
        * Linear altitude change
        * Small angle approximations
        * Quasi-steady flight

    Returns
    -------
    None
        Updates segment conditions directly:

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """        
    
    # unpack
    alt0       = segment.altitude_start 
    altf       = segment.altitude_end
    xf         = segment.distance
    air_speed  = segment.air_speed    
    beta       = segment.sideslip_angle    
    conditions = segment.state.conditions 

    # check for initial velocity
    if air_speed is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
     
    climb_angle  = np.arctan((altf-alt0)/xf)
    v_x          = np.cos(beta)*np.cos(climb_angle)*air_speed
    v_y          = np.sin(beta)*np.cos(climb_angle)*air_speed
    v_z          = np.sin(climb_angle)*air_speed 
    t_nondim     = segment.state.numerics.dimensionless.control_points
    
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0    
    
    # pack
    conditions.freestream.altitude[:,0]             = alt[:,0]
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down 
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = -v_z  